# -*- coding: utf-8 -*-

import contextlib
import datetime
from dateutil.relativedelta import relativedelta
from enum import IntEnum
import hashlib
import hmac
import io
import logging
import os
import time
import json
import re
import requests
import uuid
import random
import werkzeug.urls
import werkzeug.utils

from werkzeug.urls import Href, url_encode

from collections import defaultdict

import odoo
from odoo import api, fields
from odoo import SUPERUSER_ID
from odoo.modules.registry import Registry
from odoo.exceptions import AccessDenied, AccessError
from odoo.osv import expression
from odoo.tools.parse_version import parse_version
from odoo.tools import email_domain_extract, email_domain_normalize

from odoo.addons.auth_oauth.controllers.main import fragment_to_query_string

from odoo.addons.openerp_enterprise.models.openerp_enterprise import InvalidTagsError

from odoo.addons.saas_worker.config import saas_config
from odoo.addons.saas_worker.auth_util import SaasAuth, sign_params, tokenize_param, detokenize_param
from odoo.addons.saas_worker.compat import db_monodb, route, http, request
from odoo.addons.saas_worker.const import IAP_WS_TOKEN, IAP_WS_URL, MASTER, SAAS_BASE_URL, SCHEME, SPECIAL_PREFIXES, TEST, TIMEOUT, WEBSITE
from odoo.addons.saas_worker.trials import Frozen
from odoo.addons.saas_worker.util import ip2coordinates, from_role, get_server_names_by_ip, get_roles_from_ip, \
    humanize_version, normalize_ip_address, recaptcha_v3_get_score, SupportLevel as SL, get_edu_eol_date

from odoo.addons.sale_subscription.controllers.portal import CustomerPortal as SaleSubscriptionPortal
from odoo.addons.sale_project.controllers.portal import SaleProjectCustomerPortal

from odoo.addons.portal.controllers.portal import CustomerPortal
from odoo.addons.website.controllers.form import WebsiteForm
from odoo.addons.payment.controllers.portal import PaymentPortal
from odoo.addons.mail.controllers.mail import MailController

from odoo import _

_logger = logging.getLogger(__name__)

KEY_PATH = os.path.expanduser('~/.ssh/server_id_rsa')
PUBLISHER_WARRANTY_DB = 'openerp'
SUPPORT_PAGER = 10

class DbActivationSteps(IntEnum):
    MUG_DEAL = 2
    INVITE_PAGE = 3
    INVITED = 4
    REDEEM_GIFT = 5
    GIFT_REDEEMED = 6
    TRACK_ORDER = 7

class PublisherWarrantyController(http.Controller):
    _cp_path = '/publisher-warranty'

    @route(['/publisher-warranty'], auth='none', csrf=False, save_session=False, readonly=False)
    def index(self, action="", arg0=""):
        # TEST URL:
        #  /publisher-warranty?action=update&arg0={"dbuuid":"a","dbname":"b","language":"en","version":"7.0"}
        req = request
        result = ''
        if action == "update":
            ip_address = req.httprequest.environ['REMOTE_ADDR']
            registry = Registry(PUBLISHER_WARRANTY_DB)
            with registry.cursor() as cr:
                env = api.Environment(cr, SUPERUSER_ID, {})
                result = env['openerp.enterprise.ping'].ping(arg0, ip_address)
        return result

    @route('/publisher-warranty/unlink', methods=['GET'], auth='public', website=True)
    def unlink_database(self, token, db_uuid):
        params = {}
        params['database'] = request.env['openerp.enterprise.database'].sudo().search([('db_uuid', '=', db_uuid)])

        if params['database'] and params['database'].subscription_id:
            if params['database']._check_unlink_token(token):
                params['database']._unlink_subscription()
                params['database_unlinked'] = True
            else:
                params['invalid_link'] = True

        return request.render('openerp_enterprise.publisher_warranty_unlink', params)

    @http.route('/publisher-warranty/send-unlink-mail/json/<int:subscription_id>/<path:unlink_token>', auth='none', type='json', readonly=False)
    def odoo_enterprise_send_subscription_unlink_mail_json(self, subscription_id, unlink_token):
        subscription = request.env['sale.order'].sudo().browse(subscription_id).exists()
        # check if the subscription exists
        if not subscription:
            return {'result': False, 'reason': _('Invalid subscription')}
        # check if the security token is valid
        if not subscription._check_unlink_mail_token(unlink_token):
            # do not return a specific error message
            return {'result': False, 'reason': _('Invalid subscription')}
        # check if the order is a subscription
        if not subscription.is_subscription:
            return {'result': False, 'reason': _('Invalid subscription')}
        try:
            subscription._send_unlink_mail()
        except Exception:
            return {'result': False, 'reason': _('Email failure')}
        return {'result': True}

    @http.route('/publisher-warranty/send-unlink-mail/<int:subscription_id>/<path:unlink_token>', auth='public', website=True)
    def odoo_enterprise_send_subscription_unlink_mail(self, subscription_id, unlink_token):
        render_values = self.odoo_enterprise_send_subscription_unlink_mail_json(subscription_id, unlink_token)
        return request.render('openerp_enterprise.send_subscription_unlink_email_confirmation', render_values)

def get_location(req, ip_address=None):
    """return location from odoo request 'req'"""
    country_id = None
    location = {}
    if not ip_address:
        ip_address = req.httprequest.environ['REMOTE_ADDR']
    if ip_address:
        coordinates = ip2coordinates(ip_address)
        country_code = coordinates['country_code']
        country = req.env['res.country'].sudo().search([('code', '=', country_code.upper())], limit=1)
        country_id = country.id
        if coordinates.get('city'):
            location['city'] = coordinates['city']
        if coordinates.get('longitude'):
            location['partner_longitude'] = float(coordinates['longitude'])
        if coordinates.get('latitude'):
            location['partner_latitude'] = float(coordinates['latitude'])
        if coordinates.get('region'):
            state = country.state_ids.filtered(lambda x: x.code == coordinates.get('region'))
            location['state_id'] = state and state[0].id
    return country_id, location


def admin_check(db_name, uid):
    try:
        data = {'db_name': db_name, 'uid': uid}
        url = MASTER + '/saas_master/admin-check'
        resp = requests.get(url, params=data, auth=SaasAuth(), verify=not TEST)
        resp.raise_for_status()
    except Exception:
        _logger.exception("admin_check failed")
        return False
    return True

def build_activation_confirm_link(db_name, confirm_token, username, email, country_id):
    confirm_link_params = url_encode({
        'username': username,
        'email': email,
        'country_id': country_id})
    confirm_link = WEBSITE + "/odoo-enterprise/confirm-database/%s/%s?%s"
    return confirm_link % (db_name, confirm_token, confirm_link_params)

def get_anonymized_activation_email(database):
    # joe@mydomain.com -> j***@m***.com
    # retrieve the company email, if empty check on the first ping
    company_email = database.company_email or database.ping_ids[-1:].company_email or ''
    return company_email and re.sub(r'(.).*@(.).*(\.)', r'\1***@\2***.', company_email) or False

def send_activation_email(email, country_id, db_name, confirm_token, force_send=False):
    template_confirm_trial = request.env.ref('openerp_enterprise.confirm_trial')

    # We don't have the username anymore, so we replace it by the email address
    confirm_link = build_activation_confirm_link(db_name, confirm_token, email, email, country_id)
    context = {
        'email_to': email,
        'db_name': '%s.odoo.com' % db_name,
        'confirm_link': confirm_link,
    }
    template_confirm_trial.sudo().with_context(context).send_mail(0, force_send=force_send)

def get_extra_app_dependencies(apps, version):
    app_ids = request.env['openerp.enterprise.database.module'].sudo().search([
        ('technical_name', 'in', apps),
        ('version_id', '=', version),
    ])
    dependency_ids = app_ids.get_all_dependencies()
    new_dependencies = dependency_ids - app_ids
    return new_dependencies.mapped('technical_name')


class ContactController(WebsiteForm):

    @route('/odoo-enterprise/submit-trial', auth='public', methods=['POST'], website=True)
    def odoo_trial_submit(self, **kw):
        """ Handle the trial form
            Create the database and send the confirmation mail
            with a link to set the user password
        """
        Database = request.env['openerp.enterprise.database']
        Subscription = request.env['sale.order']
        Country = request.env['res.country']
        TrialLog = request.env['openerp.enterprise.trial.log']
        TrialRulesSudo = request.env['openerp.enterprise.trial.rules'].sudo()

        username = kw['username']
        email = kw['email'].lower().strip()
        company_name = kw['company_name']
        db_name = kw['db_name']
        company_size = kw.get('company_size')
        app = kw['selected_app']
        phone = kw['phone']
        plan = kw.get('plan')
        country_id = int(kw['country_id'])
        country_code = Country.sudo().browse(country_id).code
        contract_uuid = kw.get('cuuid', False)
        lang = kw.get('lang', 'en_US')
        hosting = kw.get('hosting')
        template = kw.get('template')
        analytics_opt_in = bool(kw.get('analytics_opt_in', False))
        verification = kw.get('verification') # verification code to bypass antispam
        lts = kw['lts'] == '1'
        ws_params = kw.get('ws_params')

        recaptcha_required = TrialLog._recaptcha_check_required()
        bypass_antispam = verification and TrialLog._verify_antispam_bypass_code(verification)

        if not TEST and not bypass_antispam:
            recaptcha_min_score = TrialLog._recaptcha_minimal_score()
            recaptcha_v3_token = kw.get('grc3_token')

            if recaptcha_v3_token is None:
                _logger.error("Odoo trial submit: recaptcha v3 token not sent from IP %s, rejected", request.httprequest.remote_addr)
                request.session.last_trial_error = -1
                return request.redirect('/odoo-enterprise/trial_error')

            recaptcha_v3_score = recaptcha_v3_get_score(
                                    saas_config['accounts']['recaptcha_v3_secretkey'],
                                    recaptcha_v3_token
                                )

            if recaptcha_v3_score != 'timeout' and (recaptcha_v3_score is False or (recaptcha_v3_score < recaptcha_min_score and recaptcha_required)):
                _logger.error("Odoo trial submit: recaptcha v3 score %r for IP %s, rejected", recaptcha_v3_score, request.httprequest.remote_addr)
                request.session.last_trial_error = -2
                return request.redirect('/odoo-enterprise/trial_error')

        subscription = False
        if contract_uuid:
            # Check that the contract exists in the database
            check_domain = [('is_subscription', '=', True), ('access_token', '=', contract_uuid)]
            subscription = Subscription.sudo().search(check_domain, limit=1)

            if not subscription:
                return request.redirect('/')

            # Check that the contract is not already attached
            check_domain = [('subscription_id', '=', subscription.id)]
            already_attached = Database.sudo().search_count(check_domain)

            if already_attached:
                return request.redirect('/odoo-enterprise/contract-already-linked')

        # anti-spam
        if not bypass_antispam:
            res = TrialRulesSudo._process_trial_creation(**kw)
            if isinstance(res, TrialRulesSudo.__class__):
                request.session.last_trial_error = res.id
                return request.redirect('/odoo-enterprise/trial_error')

            if not TrialRulesSudo._process_trial_limit(**kw):
                request.session.last_trial_error = -3
                return request.redirect('/odoo-enterprise/trial_error')

        # Check the database name
        resp = requests.get(MASTER + '/saas_master/trial_dbname/' + db_name, auth=SaasAuth(), verify=not TEST)
        db_name = resp.text

        # Create openerp_enterprise_database
        dbuuid = str(uuid.uuid4())
        confirm_token = str(uuid.uuid4())

        values = {
            'company_name': company_name,
            'db_name': db_name,
            'db_uuid': dbuuid,
            'mode': 'saas confirmed' if subscription else 'saas trial15days',
            'confirm_token': confirm_token,
            'subscription_id': subscription.id if contract_uuid else False
        }
        database = Database.sudo().create(values)

        # create dns
        version = None
        if template:
            Template = request.env['openerp.enterprise.database.template']
            db_template = Template.sudo().search([
                ('published_for', '!=', False),
                ('db_name', '=', template),
                ('last_ping', '>', fields.Date.subtract(fields.Date.today(), days=30))
            ], order="last_ping desc", limit=1)
            if not db_template:
                raise werkzeug.exceptions.NotFound()

            if db_template.published_for == 'student':
                if not kw.get('template_access_token') or not db_template._verify_template_token(kw['template_access_token']):
                    raise werkzeug.exceptions.NotFound()

            if db_template.version:
                version = db_template.version.replace('~', '-')

        if lts:
            version = request.env['ir.config_parameter'].sudo().get_param('openerp_enterprise.lts_version')

        # website scraper
        iap_ws_uuid = ''
        iap_ws_target_url = ''
        if ws_params:
            try:
                iap_ws_params = json.loads(ws_params)
                iap_ws_target_url = iap_ws_params.get('url')
                iap_ws_job = {
                    'dbuuid': dbuuid,
                    'page_count': 10,
                    'token': IAP_WS_TOKEN,
                    'url': iap_ws_target_url,
                }
                _logger.info("IAP web scraper call. dbuuid: %s. url: %s", dbuuid, iap_ws_target_url)
                iap_ws_endpoint = f'{IAP_WS_URL}/website_scraper/1.0.0/scrape'
                iap_ws_req = requests.post(iap_ws_endpoint, json={'params': iap_ws_job}, timeout=5)
                iap_ws_res = iap_ws_req.json().get('result', {})
                iap_ws_uuid = iap_ws_res.get('uuid', '')
            except Exception:
                _logger.info("Error during IAP web scraper call")

        # TODO build the trial url on accounts
        url_params = {
            'email': email,
            'dbname': db_name,
            'app': app,
            'dbuuid': dbuuid,
            'company_name': company_name,
            'country_code': country_code,
            'sync_after_creation': bool(subscription),
            'lang': lang,
            'hosting': hosting,
            'template': template,
            'version': version,
            'username': username,
            'confirm_token': tokenize_param(confirm_token),
            'spam_check': not bypass_antispam,
        }
        headers = {'X-Odoo-GeoIP': request.httprequest.environ['REMOTE_ADDR']}
        resp = requests.post(MASTER + '/saas_master/book-dns',
                             data=url_params, verify=not TEST, headers=headers,
                             allow_redirects=False, auth=SaasAuth(), timeout=TIMEOUT)
        url = resp.headers.get('location', '/')
        hosting_assigned = resp.headers.get('X-Odoo-Hosting', False)
        worker_version = resp.headers.get('X-Odoo-Version', False)
        lang_id = request.env['res.lang'].sudo().search([('code', '=', lang)]).id
        request.env['openerp.enterprise.trial.log'].sudo().create({
            'database_id': database.id,
            'creation_date': fields.Datetime.now(),
            'hosting': hosting_assigned,
            'app_list': ':' + app + ':',
            'lang_id': lang_id,
            'country_id': country_id,
            'analytics_opt_in': analytics_opt_in,
            'ip_address': request.httprequest.remote_addr,
        })

        def add_re_parameter(url, params):
            url_parsed = werkzeug.urls.url_parse(url)
            url_query = url_parsed.decode_query()
            re_url = url_query['re']
            re_url_parsed = werkzeug.urls.url_parse(re_url)
            re_url_query = re_url_parsed.decode_query()
            re_url_query = re_url_query.to_dict()  # We want only one value per key
            re_url_query.update(params)
            re_qs = url_encode(re_url_query)
            re_url = re_url_parsed.replace(query=re_qs).to_url()
            url_query['re'] = re_url
            qs = url_encode(url_query)
            return url_parsed.replace(query=qs).to_url()

        def sign_re_url(url):
            url_parsed = werkzeug.urls.url_parse(url)
            url_query = url_parsed.decode_query()
            re_url = werkzeug.urls.url_parse(url_query['re'])
            signed_params = sign_params(re_url.path, werkzeug.urls.url_decode(re_url.query))
            return add_re_parameter(url, signed_params)

        dep_modules = get_extra_app_dependencies(app.split(':'), worker_version)
        if dep_modules:
            params = {'dep_modules': ':'.join(dep_modules)}
            if analytics_opt_in and not TEST:
                # Add a flag that will instruct `trial` to install the analytics tracker module.
                params['analytics_opt_in'] = 1
            url = add_re_parameter(url, params)

        if template and kw.get('template_access_token'):
            url = add_re_parameter(url, {'duplicate_admin': True})

        if iap_ws_uuid:
            url = add_re_parameter(url, {'iap_ws_uuid': iap_ws_uuid})
            url = add_re_parameter(url, {'iap_ws_target_url': iap_ws_target_url})

        if TEST:
            # stay on same domain to allow trials on next
            url = werkzeug.urls.url_parse(url).replace(scheme='', netloc='').to_url()

        url = sign_re_url(url)

        # Create lead
        lead_args = {
            'name': db_name,
            'description': '**db: %s\n**app: %s' % (db_name, app),
            'email_from': email,
            'force_scope': 'trial',
            'plan': plan,
            'country_id': country_id,
            'contact_name': username,
            'phone': phone,
            'partner_name': company_name,
            'context': {'all_lead_kwargs': {'company_size': company_size, 'lang': lang}},
        }
        model_record = request.env['ir.model'].sudo().search([('model', '=', 'crm.lead')])
        self.insert_record(request, model_record, lead_args, {})
        return request.redirect(url, code=302, local=False)

    @route('/odoo-enterprise/trial_error', auth='public', website=True, sitemap=False, readonly=True)
    def odoo_trial_error(self, **kw):
        error_code = kw.get('err') or request.session.last_trial_error or ''

        # generate dummy request ID to integrate error code: e.g. 91d5ed0-b8121440CODE-ed0c1d5f
        # works for both positive (rule ID) and negative (fixed) error codes
        u = str(uuid.uuid4())
        req_id = '%s-%s0%s-%s' % (u[-12:-5], u[:7], error_code, u[-8:])

        params = {
            'timestamp': datetime.datetime.now().replace(microsecond=0),
            'req_id': req_id,
            'random': random.randint(0, 2), # for dynamic bg
        }
        _logger.info(f"Trial error page: ERR {error_code} - REQ-ID {req_id} - IP {request.httprequest.remote_addr}")
        return request.render("openerp_enterprise.trial_creation_error", params)

class EnterpriseController(PaymentPortal):

    @route('/odoo-enterprise/databases', auth='user', website=True, type='json', readonly=True)
    def get_databases(self, **kw):
        Database = request.env['openerp.enterprise.database']
        all_instances = [db for db in Database._get_instances(alive=True) if db['is_saas']]
        restricted_instance_ids = set()
        if request.session.impostor:
            all_instance_ids = [db['db_id'] for db in all_instances]
            restricted_instance_ids = set(Database.browse(all_instance_ids).sudo()._filter_restricted(request.session.impostor).ids)

        def set_buttons(db):
            dropped = db['visibility'] == 'hidden'
            base_rule = not dropped and (db['is_admin'] or db['is_edu'])
            db['btns'] = {
                'dropped': dropped,
                'duplicate': base_rule,
                'rename': base_rule,
                'backup': base_rule and (db['total_size'] < 10 * 1024**3),
                'domains': db['is_saas'] and db['is_admin'] and not db['is_edu'] and not dropped,
                'share': db['is_admin'] and not dropped,
                'delete': not bool(db['is_paid']) and db['is_admin'] and not dropped,
                'invite': base_rule,
                'remove_users': base_rule,
                'tags': base_rule,
            }

        server_names_by_ip = get_server_names_by_ip()
        for db in all_instances:
            set_buttons(db)
            # if database is restricted, disable all actions but tags
            if db['db_id'] in restricted_instance_ids:
                db['btns'].update(dict.fromkeys(set(db['btns']) - {'tags'}, False))

            # remove information needed for buttons but not for UI
            db.pop('total_size', None)

            # retrieve the hosting server
            oedb = Database.sudo().browse(db['db_id'])
            last_ip = oedb.last_ip
            server_fqdn = server_names_by_ip.get(normalize_ip_address(last_ip), False)
            if server_fqdn:
                version = humanize_version(oedb.version)
                db['server'] = server_fqdn
                db['server_port'] = odoo.addons.saas_worker.metabase.get_port(version)
            else:
                _logger.info("get_databases: no FQDN found for ip=%r (dbid=%r, dbname=%r)", last_ip, db['db_id'], db['db_name'])

        allowed_visiblity = ['customer']
        is_support = bool(request.session.impostor or request.env.user.has_group('base.group_user'))
        if is_support: # if the user is an employee (or is spoofing someone)
            allowed_visiblity.append('support')
            allowed_visiblity.append('hidden')

        all_instances = [db for db in all_instances if db['visibility'] in allowed_visiblity]
        return {
            'databases': all_instances,
        }

    @route(['/my/databases', '/my/databases/manage', '/account'], auth='public', website=True, sitemap=False, readonly=True)
    def show_instances(self, **kw):
        if request.httprequest.path == '/account':
            return request.redirect('/my/home', 301)
        if request.env.user._is_public() and 'need_auth' in kw:
            return_url = werkzeug.urls.url_quote(request.httprequest.url)
            return request.redirect('/web/login?redirect={}'.format(return_url))
        comm_partner = request.env.user.commercial_partner_id
        params = {
            'share': 'share' in kw,
            'support': bool(request.session.impostor or request.env.user.has_group('base.group_user')),
            'error_msg': kw.get('error_msg', ''),
            'email': request.env.user.partner_id.email if request.session.uid else '',
            'is_accounting_firm': comm_partner.portal_accounting_firm_level and comm_partner.is_published,
        }
        return request.render("openerp_enterprise.databases", params)

    @route('/my/databases/check', auth='user', website=True, type='json', readonly=True)
    def database_check(self, db_ids):
        """
        Returns a mapping {id: bool}
        telling whether the current user is an admin on the database identified by id,
        for each id of db_ids
        """
        dbs = request.env['openerp.enterprise.database'].sudo().browse(db_ids)
        return {
            db.id: db._is_user_admin(request.env.user)
            for db in dbs
        }

    @route('/my/databases/connect/<string:db_uuid>', auth='user', website=True)
    def database_connect(self, db_uuid):
        """Show a self-validating form that will POST to the given database"""
        db = request.env['openerp.enterprise.database'].sudo().search([('db_uuid', '=', db_uuid)], limit=1)
        if not db:
            raise werkzeug.exceptions.NotFound()

        if request.session.impostor:
            reason = f"db: {db.db_name}"
            spoof_vals = {
                'ip': request.httprequest.environ['REMOTE_ADDR'],
                'user_id': request.session.impostor,
                'reason': reason,
                'remote_uid': request.session.uid,
                'remote_login': request.session.login,
            }
            access = db._check_and_log_spoof('connect_as', **spoof_vals)
            if not access:
                return request.redirect(
                    '/my/databases?%s' % (werkzeug.urls.url_encode({
                        'error_msg': _('Error: This operation is restricted. You need to ask for a staff access authorization to the customer.'),
                    }))
                )

        token = request.env['auth.oauth2.token'].sudo()._get_trial_token(request.uid, db_uuid)
        ip = request.httprequest.environ['REMOTE_ADDR']
        _logger.info('new-token: issued token=%s uid=%d db_uuid=%s ip=%s', token[:8], request.uid, db_uuid, ip)

        params = dict(
            db_name=db.db_name,
            db_url=db.url,
            db_uuid=db_uuid,
            token=token,
        )
        return request.render("openerp_enterprise.database_connect", params)

    @route('/my/databases/delete', auth='user', website=True, methods=['POST'])
    def database_delete(self, dbid, **kwargs):
        # we retrieve the db instance
        db = request.env['openerp.enterprise.database'].sudo().browse(int(dbid))

        if request.session.impostor:
            reason = f"db: {db.db_name}"
            spoof_vals = {
                'ip': request.httprequest.environ['REMOTE_ADDR'],
                'user_id': request.session.impostor,
                'reason': reason,
                'remote_uid': request.session.uid,
                'remote_login': request.session.login,
            }
            access = db._check_and_log_spoof('drop', **spoof_vals)
            if not access:
                return request.redirect(
                    '/my/databases?%s' % (werkzeug.urls.url_encode({
                        'error_msg': _('Error: This operation is restricted. You need to ask for a staff access authorization to the customer.'),
                    }))
                )

        # This check is deactivated, the 12.0 databases are not compatible with our way of deciding who's the oauth admin
        # Only the owner of the db can delete it
        if not db._is_user_admin(request.env.user):
            error_msg = url_encode({'error_msg': "You must be the administrator of the database to delete it!"})
            return request.redirect('/my/databases?%s' % error_msg)
        result = {'result': False, 'message': ''}
        # call the 'drop' route on the trial server
        try:
            server_names_by_ip = get_server_names_by_ip()
            server_fqdn = server_names_by_ip[normalize_ip_address(db.last_ip)]
            db_drop_url = '%s://%s/saas_worker/drop' % (SCHEME, server_fqdn)
            json_params = {'params': {'db_name': db.db_name, 'db_uuid': db.db_uuid, 'db_admin_oauth_uid': request.uid}}
            response = requests.post(db_drop_url, json=json_params, auth=SaasAuth(), timeout=60)  # 60s timeout i/o 15 from const.py
            response.raise_for_status()
            result = response.json().get('result', {})
        except Exception:
            result.update(result=False, message='Error: We are not able to delete your database now, please retry later.')
        # if the db has been successfully marked as deleted, then
        if result.get('result') is True:
            # we need to commit transactions because we receive a ping from trial which write in openerp.enterprise.database
            # and cause a concurrent transaction
            request.env.cr.commit()
        if result.get('result') is True or result.get('message') == 'Unknown database.':
            # we hide the db so the user doesn't see it on the databases list
            db.sudo().write({'visibility': 'hidden'})
            # and redirect the user to the dabatases manage page where the db should have disappeared
            return request.redirect('/my/databases')
        # well there was an error
        error_msg = url_encode({'error_msg': result.get('message', "Error while deleting your database")})
        return request.redirect('/my/databases?%s' % error_msg)

    @route('/my/databases/send-activation-email', auth='user', type='json')
    def database_send_activation_email(self, db_uuid, email):
        oauth_user_id = request.env['openerp.enterprise.oauth.user'].sudo().search([
            ('user_id', '=', request.session.uid),
            ('is_admin', '=', True),
            ('database_id.db_uuid', '=', db_uuid),
        ])
        if not oauth_user_id:
            return {'result': False}

        database = oauth_user_id.database_id
        if database.is_activated():
            return {'result': False}

        country = database.company_country
        geoip_country_id, dummy = get_location(request)
        confirm_token = tokenize_param(database.confirm_token)
        send_activation_email(email, country.id or geoip_country_id, database.db_name, confirm_token, force_send=True)

        return {'result': True}

    @route('/my/databases/set_tags_multidb', auth='user', website=True, type='json')
    def database_set_tags_multidb(self, databases, tag_changes):
        db_uuids = [db['db_uuid'] for db in databases]
        oauth_user_ids = request.env['openerp.enterprise.oauth.user'].sudo().search([
            ('user_id', '=', request.session.uid),
            ('is_admin', '=', True),
            ('database_id.db_uuid', 'in', db_uuids),
        ])
        if len(oauth_user_ids) != len(databases):
            return {
                'result': False,
                'error_code': 'no_access',
                'error_data': list(set(db['db_name'] for db in databases) - set(u.database_id.db_name for u in oauth_user_ids)),
            }

        try:
            oauth_user_ids.mapped('database_id').update_tags(tag_changes)
            return {'result': True, 'tags': {u.database_id.db_uuid: u.database_id.tag_ids.mapped('name') for u in oauth_user_ids}}
        except InvalidTagsError as e:
            return {'result': False, 'error_code': 'invalid_tags', 'error_data': e.invalid_tags}

    @route('/my/databases/autocomplete-tags', auth='user', website=True, type='json', readonly=True)
    def database_autocomplete_tags(self):
        """
        Retrieves the list of all the tags that are set on at least one of a user's databases, for autocompletion feature.
        """
        Database = request.env['openerp.enterprise.database'].sudo()
        DatabaseTag = request.env['openerp.enterprise.database.tag'].sudo()

        user_tag_ids = Database.search([('registered_user_ids.user_id', '=', request.env.user.id)]).tag_ids
        tags = DatabaseTag.search([('id', 'in', user_tag_ids.ids)], order='name ASC').mapped('name')
        return tags

    @route([
        '/my/database/staff_log/<int:db_id>',
        '/my/database/staff_log/<int:db_id>/page/<int:page>'
    ], auth='user', page=0, website=True, methods=['GET'], readonly=True)
    def staff_log(self, db_id, page=0):
        database = request.env['openerp.enterprise.database'].sudo().browse(db_id)
        database_users = database.registered_user_ids.mapped('user_id')
        if request.env.user in database_users or request.env.user.has_group('base.group_user'):
            spoof_count = database._get_spoof_count()
            url = '/my/database/staff_log/%s' % str(database.id)
            pager = request.website.pager(
                url=url, total=spoof_count, page=page, step=SUPPORT_PAGER, scope=10,
                url_args={}
            )
            user_is_staff = request.env.user.has_group('base.group_user') or request.session.impostor
            spoof_logs = database._get_spoof_lines(staff=user_is_staff, offset=pager['offset'], limit=SUPPORT_PAGER)
            subscription = database.subscription_id or database.parent_id.subscription_id
            return request.render('openerp_enterprise.portal_database_support_logging', {
                'user_is_staff': user_is_staff,
                'pager': pager,
                'support_access': spoof_logs,
                'account': subscription,
                'database': database,
            })
        else:
            raise werkzeug.exceptions.Forbidden()

    @route('/odoo-enterprise/in-my-accounting-firm', type='json', auth='user', readonly=True)
    def same_accounting_firm(self, emails):
        """ return whether the emails can be considered as being in the same accounting firm as the current user """
        current_user = request.env.user
        if not current_user.commercial_partner_id.portal_accounting_firm_level:
            return False
        current_domain = current_user.login.partition('@')[2].lower()

        other_users = set(u.email
                          for u in request.env['res.users'].search([('login', 'in', emails)])
                          if u.commercial_partner_id == current_user.commercial_partner_id)
        return {email: (email in other_users
                        or email.partition('@')[2].lower() == current_domain)
                for email in emails}

    def _buy_get_country(self):
        """
        Return the country for the self buying process.

        Take the country from GeoIP if user is not logged; otherwise use partner's country. If the
        user is an Odoo employee, allow spoofing locations.
        """
        country_code = request.geoip.country_code or 'BE'
        country = request.env['res.country'].search([('code', '=', country_code)])
        if request.website.partner_id != request.env.user.partner_id:
            partner_id = request.env.user.partner_id
            # allow localization spoofing for employees only
            country = country if request.env.user.has_group('base.group_user') else partner_id.country_id
        return country

    @route(['/odoo-enterprise/upgrade', '/buy', '/upgrade'], auth='public', website=True, sitemap=True, readonly=True)
    def odoo_upgrade(self, **kw):
        kw['hosting'] = 'on_premise'
        return request.redirect('/pricing#{}'.format(url_encode(kw)))

    @route(['/odoo-enterprise/renew'], auth='public', website=True, sitemap=False, readonly=True)
    def odoo_renew(self, **kw):
        """ Render renewal form for Enterprise/Online Yearly subscriptions.

        This route exepect a 'contract' GET param containing a subscription code
        that will then be browsed in sudo to display a renewal form with the
        current information we have on the subscription. This form will contain
        two payment providers: Wire Payment (always, Odoo S.A.) and either Ogone US
        (for Odoo Inc. subs) or Ogone BE (everyone else).

        Also, notice that the route is public: we assume the contract code is
        enough obfuscation. This makes it easy for customer coming from their
        on-premise database to pay directly.
        """
        code = kw.get('contract')

        contract_sudo = request.env['sale.order'].sudo()._get_most_valid_subscription(code)
        if contract_sudo.subscription_state == '2_renewal' and contract_sudo.state != 'sent':
            # parent subscription may have churn but the renewal is not yet
            # validated, so we redirect to parent subscription anyway
            contract_sudo = contract_sudo.subscription_id
        if contract_sudo.subscription_state in ('3_progress', '4_paused', '6_churn'):
            sent_renewals = contract_sudo.subscription_child_ids.filtered(
                lambda x: x.subscription_state == '2_renewal' and x.state == 'sent'
            )
            if sent_renewals:
                # a renewal was already sent to the customer, redirect to that renewal
                contract_sudo = sent_renewals[0]
        if not contract_sudo:
            raise request.not_found()
        try:
            self._document_check_access('sale.order', contract_sudo.id, False)
            return request.redirect(contract_sudo.access_url)
        except AccessError:
            # the user cannot access the subscription, ask him to logged in
            kw = url_encode({'redirect': contract_sudo.access_url})
            return request.redirect(f"/web/login?{kw}")

    @route(['/odoo-enterprise/upsell'], auth='public', website=True, sitemap=False)
    def odoo_upsell(self, contract=None, **kw):
        """Handle the upsell mechanism for subscriptions.

        Check if a new upsell is needed.
        If that's the case, the system will generate the upsell and redirect to the quote
        transparently; if not, either an old upsell is still valid (then you are redirected as
        well) or none is necessary, in which case there's nothing here (404).

        :param str contract: a valid subscription code (Mxxxxxxxxxxx)
        """
        if contract:
            sub = request.env['sale.order'].sudo()._get_most_valid_subscription(contract)
            if sub:
                order_id = sub.enterprise_upsell()
                order = sub.env['sale.order'].sudo().browse(order_id)
                if order_id:
                    quote_url = order.get_portal_url()
                    return request.redirect(quote_url)
        return request.render('openerp_enterprise.oops_upsell')

    @route('/trial', auth='public', website=True, sitemap=True, readonly=True)
    def odoo_trial(self, **kw):
        session_uid = request.session.uid

        countries = request.env['res.country'].sudo().search([])
        geoip_country_id, location = get_location(request)

        company_name = kw.get('company_name', '')
        partner = request.env.user.partner_id.sudo().read(['email', 'display_name', 'phone'])

        # Languages proposed in the subscription form
        # We keep only the languages which has been translated over 60%
        # See https://www.transifex.com/odoo/odoo-14/
        # Languages are hardcoded because we just want the native language
        # without the english version and the related country
        proposed_languages = [
            # (code, url_code, name)
            ('ar_001', 'ar_001', 'الْعَرَبيّة'),
            ('cs_CZ', 'cs_CZ', 'Čeština'),
            ('da_DK', 'da_DK', 'Dansk'),
            ('de_DE', 'de_DE', 'Deutsch'),
            ('en_US', 'en_US', 'English'),
            ('es_ES', 'es_ES', 'Español'),
            ('es_419', 'es_419', 'Español (América Latina)'),
            ('et_EE', 'et_EE', 'Eesti keel'),
            ('fr_FR', 'fr_FR', 'Français'),
            ('he_IL', 'he_IL', 'עִבְרִי'),
            ('it_IT', 'it_IT', 'Italiano'),
            ('ja_JP', 'ja_JP', '日本語'),
            ('ko_KR', 'ko_KR', '한국어'),
            ('mn_MN', 'mn_MN', 'монгол'),
            ('nl_NL', 'nl_NL', 'Nederlands'),
            ('pl_PL', 'pl_PL', 'Język polski'),
            ('pt_BR', 'pt_BR', 'Português (BR)'),
            ('ru_RU', 'ru_RU', 'Русский язык'),
            ('ro_RO', 'ro_RO', 'română'),
            ('sk_SK', 'sk_SK', 'Slovenský'),
            ('tr_TR', 'tr_TR', 'Türkçe'),
            ('uk_UA', 'uk_UA', 'українська'),
            ('vi_VN', 'vi_VN', 'Tiếng Việt'),
            ('zh_CN', 'zh_CN', '简体中文'),
            ('zh_TW', 'zh_TW', '繁體中文'),
        ]

        # Expressed by a number between 0 and 100.
        probability = int(request.env['ir.config_parameter'].sudo().get_param('saas_fullstory.sampling_probability', default=0))

        # Need a website pricelist to exclude paying apps from old pricing
        # per app / par user (e.g. apps hr) that are free now
        paid_modules = request.env['openerp.enterprise.database.module'].sudo().pricelist2modules_and_deps(
            kw.get('lts') and request.env['ir.config_parameter'].sudo().get_param('openerp_enterprise.lts_version')
        )

        params = {
            'countries': countries,
            'geoip_country_id': geoip_country_id,
            'company_name': company_name,
            'selected_app': kw.get('selected_app', ''),
            'partner': session_uid and partner and partner[0] or {},
            'cuuid': kw.get('cuuid', ''),
            'lang': kw.get('lang', request.lang.code),
            'proposed_languages': proposed_languages,
            'hosting': kw.get('hosting'),
            'template': kw.get('template'),
            # Disabled by executive order of FP, as of 2024-02. Keeping the feature just in case
            # we need it again in the future.
            #'show_analytics_feature': not kw.get('template') and random.randint(1, 100) <= probability,
            'verification': kw.get('verification'),
            'lts': kw.get('lts', 0),
            'apps': paid_modules,
            'edu': 'edu' in kw,
            'template_access_token': kw.get('token', False),
            'ws_params': kw.get('ws_params'),
        }

        if kw.get('form_light'):
            return request.render('openerp_enterprise.trial_form_light', params)
        return request.render('openerp_enterprise.trial_form', params)

    @route('/odoo-enterprise/trial-hint', auth='public', type="json", readonly=True)
    def odoo_trial_hint(self, app_list, hosting=None, lts=False, **kw):
        data = {
            'ip_address': request.httprequest.remote_addr,
            'hosting': hosting,
        }
        if lts:
            data['version'] = request.env['ir.config_parameter'].sudo().get_param('openerp_enterprise.lts_version')

        res = requests.post(MASTER + '/saas_master/dn/hint', data=data, auth=SaasAuth(), timeout=TIMEOUT)
        res.raise_for_status()
        res_data = res.json()

        deps = get_extra_app_dependencies(app_list, res_data['worker_version'])
        res_data.update({
            'deps': deps,
            'success': True,
        })
        return res_data

    @route('/odoo-enterprise/confirm-database/<db_name>/<confirm_token>', auth='public', website=True)
    def confirm_database(self, db_name, confirm_token, **kw):
        """ The link in the confirmation mail point to this route.
            Just render the form to let the user set a password
        """
        # TODO: remove these params, redundant with db/user info based on token
        email = kw['email'].lower()
        name = kw['username']
        try:
            country_id = int(kw['country_id'])
        except ValueError:
            country_id = request.env.ref('base.be').id

        DB_sudo = request.env['openerp.enterprise.database'].sudo()
        token_is_valid = DB_sudo.search_count([('confirm_token', '=', detokenize_param(confirm_token))]) == 1
        if token_is_valid:
            user = request.env['res.users'].sudo().search([('login', '=ilike', email)])
            user_exists = (len(user) == 1)
            params = {
                'db_name': db_name,
                'confirm_token': confirm_token,
                'email': email, # login
                'username': name,
                'country_id': country_id,
            }
            if user_exists:
                # proceed directly to database validation for that user
                return self.confirm_database_submit(**params)
            else:
                # quick signup form
                return request.render('openerp_enterprise.confirm_database_form', params)
        # already validated or invalid
        return request.redirect('/database_validated')

    @route('/odoo-enterprise/submit-confirm-database', auth='public', methods=['POST'], website=True)
    def confirm_database_submit(self, **kw):
        """ Handle the password form. Create the user and attach it to the database,
            then destroy the confirm_token to invalid the confirmation link
        """
        DB_sudo = request.env['openerp.enterprise.database'].sudo()
        Users_sudo = request.env['res.users'].sudo()
        Leads_sudo = request.env['crm.lead'].sudo()

        # mandatory params
        db_name = kw['db_name']
        confirm_token = kw['confirm_token']
        email = kw['email']
        username = kw['username']
        country_id = int(kw['country_id'])

        # optional params: only provided in signup form
        password = kw.get('password')
        password_confirmation = kw.get('password-confirmation')
        if password != password_confirmation:
            return request.redirect(request.httprequest.referrer)

        # Check the confirm token
        validation_domain = [('confirm_token', '=', detokenize_param(confirm_token))]
        db_sudo = DB_sudo.search(validation_domain, limit=1)
        if not db_sudo:
            return request.redirect('/')

        # Verify user exists and is authenticated
        user = Users_sudo.with_context(active_test=False).search([('login', '=ilike', email)], limit=1)

        # if the user exists but is archived, we need to warn him that he has to contact the support team
        if user and not user.active:
            return request.render("openerp_enterprise.database_validation_user_archived")

        if user and request.session.uid != user.id:
            # force user to login, to prove ownership
            request.session.logout(keep_db=True)
            db_host = '%s.%s' % (db_name, SAAS_BASE_URL)
            confirm_url = Href('/odoo-enterprise/confirm-database/')(
                db_name, confirm_token, email=email, username=username, country_id=country_id)
            redir_url = Href('/web/login')(login=email, redirect=confirm_url,
                                           scope='activation', redirect_hostname=db_host)
            return request.redirect(redir_url)

        # Or create it and authenticate
        if not user:
            # heuristic to link the partner to its lead
            # TODO the id should go in the mail link, with an HMAC
            domain = [('email_from', '=', email), ('partner_id', '=', False)]
            lead_sudo = Leads_sudo.search(domain, limit=1, order='create_date desc')

            # Create the user
            values = {'name': lead_sudo.contact_name or username, 'login': email, 'email': email, 'password': password}
            login, _ = Users_sudo.signup(values, None)
            request.env.cr.commit()

            # Authenticate the user
            uid = request.session.authenticate(request.db, login, password)
            user = Users_sudo.browse(uid)
            user.partner_id.write({'country_id': country_id})

            # update partner with the details from the lead
            # only update fields that come from the registration form
            partner_data = lead_sudo._prepare_customer_values(lead_sudo.contact_name, False)
            partner_data['company_name'] = lead_sudo.partner_name
            user.partner_id.update({k: v for k, v in partner_data.items() if v is not False})

            # Link the lead to the newly created partner
            lead_sudo.write({'partner_id': user.partner_id.id})

        # check if the db exists on the worker and is frozen
        frozen_url = '%s://%s.%s/saas_worker/frozen/%s' % (SCHEME, db_name, SAAS_BASE_URL, db_name)
        res = requests.get(frozen_url + '/status', allow_redirects=False, auth=SaasAuth())
        if not res.ok or res.status_code != 200:
            # if the DB was frozen, but already got deleted, they will get a redirect here
            raise werkzeug.exceptions.NotFound()
        if res.json()['frozen']:
            params = url_encode({
                'redir_to': build_activation_confirm_link(db_name, confirm_token, username, email, country_id)
            })
            return request.redirect(frozen_url + '/thaw/' + db_sudo.db_uuid + '?' + params, local=False)

        # Attach the user to the database
        params = {
            'odoo_com_uid': user.id,
            'name': user.name,
            'email': user.email,
            'action': 'base.action_run_ir_action_todo',
            'activate_freemium': False,
            'confirm_token': confirm_token,
            # Redirect the user to the SaaS invitation page
            'invite': True,
            'db_uuid': db_sudo.db_uuid,
        }

        url = '%s://%s.%s/saas_worker/signin?%s' % (SCHEME, db_name, SAAS_BASE_URL, url_encode(params))

        # Invalidate the confirm token (to extract in a function)
        db_sudo.write({'mode': 'saas confirmed', 'confirm_token': False})
        request.env.cr.commit()

        try:
            # Update metabase
            sync_url = '%s://%s.%s/saas_worker/synchronize_metabase' % (SCHEME, db_name, SAAS_BASE_URL)
            sync_params = {'dbuuids': json.dumps([db_sudo.db_uuid])}
            requests.post(sync_url, data=sync_params, auth=SaasAuth(), timeout=TIMEOUT)
        except Exception:
            pass  # Don't crash for a simple synchronization

        return request.redirect(url, local=False)

    @route('/database_validated', auth='public', website=True, sitemap=False, readonly=True)
    def database_validated(self, *kw):
        return request.render("openerp_enterprise.database_validated")

    @from_role('trial')
    @route('/odoo-enterprise/log-trial-creation', methods=['POST'], type='http', auth='none', csrf=False, readonly=False)
    def update_trial_log(self, **kw):
        if kw.get('db_uuid'):
            database = request.env['openerp.enterprise.database'].sudo().search([('db_uuid', '=', kw['db_uuid'])])
            trial_log = request.env['openerp.enterprise.trial.log'].sudo().search([('database_id', '=', database.id)])
            trial_log.write({
                'version': kw.get('version'),
                'build_time': kw.get('build_time'),
                'creation_method': kw.get('creation_method'),
            })
        return

    @from_role('trial')
    @route('/odoo-enterprise/reset-database-token', methods=['POST'], type='http', auth='none', csrf=False, readonly=False)
    def route_reset_database_token(self, dbuuid, **kw):
        Database = request.env["openerp.enterprise.database"]
        database = Database.sudo().search([('db_uuid', '=', dbuuid)])

        database.write({'confirm_token': str(uuid.uuid4())})

        # on duplicate database, force scan of wheatbot for db with website installed
        # (Only called when not duplicated by support staff)
        if database.hosting == 'saas' and database._module_installed('website'):
            db_sub = database.parent_id.subscription_id
            paid_sub = db_sub and (db_sub.enterprise_is_enterprise or db_sub.enterprise_is_partnership)
            # we assume that paid databases are less likely to be used for abuse, and they have
            # survived the initial scan period. If a paid db is still used for abuse that will get
            # reported externally and we'll track it down all the way down to the paid database.
            if not paid_sub:
                request.env['openerp.enterprise.trial.log'].sudo().create({
                    'database_id': database.id,
                    'creation_date': datetime.datetime.now() - datetime.timedelta(minutes=30),
                    'hosting': '-',
                    'app_list': ':website:',
                    'country_id': database.company_country.id,
                    'creation_method': 'dup'
                })
        return

    @from_role('trial')
    @route('/odoo-enterprise/send-activation-email', type='http', auth='public', csrf=False)
    def route_send_activation_email(self, db_name, activation_email=False, country_code=False, **kw):
        Database = request.env['openerp.enterprise.database']
        database = Database.sudo().search([('db_name', '=', db_name)], order='create_date desc', limit=1)

        if database and not database.is_activated():
            confirm_token = tokenize_param(database.confirm_token)
            if country_code:
                country = request.env['res.country'].sudo().search([('code', '=', country_code)], limit=1)
            else:
                country = database.company_country
            activation_email = activation_email or database.company_email
            geoip_country_id, dummy = get_location(request)

            send_activation_email(
                activation_email, country.id or geoip_country_id,
                database.db_name, confirm_token,
                force_send=True
            )
            return json.dumps({'result': {'confirm_token': confirm_token}})

        return json.dumps({'result': False})

    @route('/odoo-enterprise/upgrade-online/<db_uuid>', auth='user', website=True)
    def upgrade_online(self, db_uuid, force_create=False, **kw):
        Database = request.env['openerp.enterprise.database']
        Subscription = request.env['sale.order']

        partner = request.env.user.partner_id
        database = Database.sudo().search([('db_uuid', '=', db_uuid)], limit=1)
        subscription = database.subscription_id

        if subscription and subscription.subscription_state in ('3_progress', '4_paused', '5_renewed', '6_churn'):  # not draft
            # redirect to contract
            url = '/my/subscription/%s/%s' % (subscription.id, subscription.access_token)
            return request.redirect(url)

        if not force_create:
            # Check if there exists contracts attached to this partner and not linked to any database
            tag_plan_id = request.env.ref('openerp_enterprise.%s_plan_tag' % database.get_plan()).id
            free_subscription_domain = [
                ('partner_id', 'in', [partner.id, partner.commercial_partner_id.id]),
                ('sale_order_template_id.subscription_tag_ids', '=', tag_plan_id),
                ('subscription_state', '=', '3_progress'), ('to_renew', '=', False),
                ('is_subscription', '=', True),
                ('database_ids', '=', False),
            ]
            free_subscriptions = Subscription.sudo().search(free_subscription_domain)

            if free_subscriptions:
                # Offer the user to choose among the free contracts
                render_params = {'analytic_accounts': free_subscriptions, 'db_uuid': db_uuid}
                return request.render('openerp_enterprise.contract_selector', render_params)

        # Create the contract
        country_code = request.geoip.country_code
        website_pl = request.website.pricelist_id
        pricelist_ids = request.website._get_pl_partner_order(country_code, False, request.env.user.partner_id.property_product_pricelist.id, False, request.env['product.pricelist'].browse([]))
        pricelist_id = pricelist_ids and pricelist_ids[0] or website_pl.id
        subscription_id = Subscription.sudo().create_odoo_online_contract({
            'db_uuid': db_uuid,
            'uid': request.env.user.id,
            'pricelist_id': pricelist_id
        })
        subscription = Subscription.sudo().browse(subscription_id)

        if not subscription.amount_total:  # freemium has been confirmed
            url = "/my/subscription/%s/%s" % (subscription.id, subscription._portal_ensure_token())
        elif subscription.state == 'draft':  # shop refuses sale.order in state 'sent'
            request.session['sale_order_id'] = subscription.id
            url = "/shop/checkout?express=1"
        else:  # shoud never be in this case
            url = "/error?%s&%s&%s" % (subscription.id, subscription.state, subscription.amount_total)

        return request.redirect(url)

    @from_role('trial')
    @route('/odoo-enterprise/link-contract-to-database-json', auth='none', type='json', readonly=False)
    def link_contract_to_database_json(self, code, db_uuid):
        Database = request.env['openerp.enterprise.database'].with_user(SUPERUSER_ID)
        Subscription = request.env['sale.order'].sudo()

        database = Database.search([('db_uuid', '=', db_uuid)])
        if not database:
            return {'success': False, 'error': 'db_not_found'}

        if database.subscription_id and database.subscription_id.subscription_state != '1_draft':
            return {'success': False, 'error': 'already_linked'}

        subscription = Subscription._get_most_valid_subscription(code)
        if not subscription or subscription.enterprise_is_partnership:
            return {'success': False, 'error': 'invalid_code'}

        is_already_attached = Database.search([('subscription_id', '=', subscription.id)])
        if is_already_attached:
            return {
                'success': False,
                'error': 'code_already_used',
                'database_already_linked_send_mail_url': subscription._get_unlink_mail_send_url(),
            }

        database.write({'subscription_id': subscription.id})

        request.env.cr.commit()
        sync_url = '%s://%s.%s/saas_worker/synchronize_metabase' % (SCHEME, database.db_name, SAAS_BASE_URL)
        sync_params = {'dbuuids': json.dumps([database.db_uuid])}
        requests.post(sync_url, data=sync_params, auth=SaasAuth(), timeout=15)
        return {'success': True}

    @route('/odoo-enterprise/link-contract-to-database', auth='user', methods=['POST'])
    def link_contract_to_database(self, cuuid, db_uuid, **kw):
        Database = request.env['openerp.enterprise.database']
        Subscription = request.env['sale.order']

        database = Database.sudo().search([('db_uuid', '=', db_uuid), ('subscription_id', '=', False)])

        contract_domain = [('access_token', '=', cuuid), ('is_subscription', '=', True)]
        subscription = Subscription.sudo().search(contract_domain)

        is_already_attached = Database.sudo().search([('subscription_id', '=', subscription.id)])
        if not database or not subscription or is_already_attached:
            return request.redirect('/')

        database.write({'subscription_id': subscription.id})

        request.env.cr.commit()
        sync_url = '%s://%s.%s/saas_worker/synchronize_metabase' % (SCHEME, database.db_name, SAAS_BASE_URL)
        sync_params = {'dbuuids': json.dumps([database.db_uuid])}
        requests.post(sync_url, data=sync_params, auth=SaasAuth(), timeout=15)

        url = '/my/subscription/%s/%s' % (subscription.id, subscription.access_token)
        return request.redirect(url)

    @route('/odoo-enterprise/github/add', type='http', auth='user', methods=['POST'])
    def add_github_user(self, github_username, sub_uuid, **kwargs):
        GithubUser = request.env['openerp.enterprise.github.user'].sudo()
        subscription = request.env['sale.order'].search([('access_token', '=', sub_uuid), ('is_subscription', '=', True)])
        if not subscription:
            raise werkzeug.exceptions.NotFound()
        if len(subscription.sudo().github_user_ids) >= subscription.sudo().github_limit:
            raise werkzeug.exceptions.Forbidden()
        existing_github_user = GithubUser.search([('name', '=', github_username)])
        if existing_github_user.status == 'none':
            # user previously disabled (dead sub) -> can be re-added
            existing_github_user.unlink()
        elif existing_github_user:
            # user still active -> block duplicate
            return request.redirect('/my/subscription/%s?github&gh_uniq=%s' % (subscription.id, github_username))
        GithubUser.create({'name': github_username, 'subscription_id': subscription.id})
        return request.redirect('/my/subscription/%s?github' % subscription.id)

    @route('/odoo-enterprise/github/remove', type='http', auth='user', methods=['POST'])
    def remove_github_user(self, github_user_id, sub_uuid, **kwargs):
        GithubUser = request.env['openerp.enterprise.github.user'].sudo()
        user = GithubUser.search([('id', '=', int(github_user_id))])
        subscription = request.env['sale.order'].search([('access_token', '=', sub_uuid), ('is_subscription', '=', True)])
        if not subscription or user.subscription_id != subscription:
            raise werkzeug.exceptions.NotFound()
        user.unlink()
        return request.redirect('/my/subscription/%s?github' % subscription.id)

    @route('/odoo-enterprise/contract-already-linked', website=True)
    def contract_already_linked(self, **kw):
        return request.render('openerp_enterprise.contract_already_linked')

    @route("/odoo-enterprise/database/open/<uuid>", type="http", auth="user", website=False, readonly=True)
    def open_db(self, uuid):
        db = request.env["openerp.enterprise.database"].search([("db_uuid", "=", uuid)], limit=1, order="db_uuid")
        if not db:
            raise werkzeug.exceptions.NotFound()
        action = request.env.ref("openerp_enterprise.action_pw_user_database_form").id
        redirect = f"/web#action={action}&model=openerp.enterprise.database&id={db.id}&view_type=form"
        return request.redirect(redirect)

    @route('/odoo-enterprise/template/<int:template_id>/update', type='http', auth='user', website=True)
    def odoo_enterprise_template_update(self, template_id, **kwargs):
        template = request.env['openerp.enterprise.database.template'].sudo().browse(template_id)
        database = template.database_id

        if not database.hosted_on_saas or not database._is_edu():
            return request.render("http_routing.403")
        is_admin = admin_check(database.db_name, request.uid)
        if not is_admin:
            return request.render("http_routing.403")

        if request.httprequest.method == 'POST':
            template = database.templatize(kwargs)
            return request.redirect('/odoo-enterprise/template/%s' % template.id)
        elif request.httprequest.method == 'GET':
            if kwargs.get('stop_sharing', None) == 'true':
                database.templatize({'published_for': False})
                return request.redirect('/odoo-enterprise/template')

            return request.render('openerp_enterprise.openerp_enterprise_database_template_share', {'template': template})

    @route('/odoo-enterprise/template/<int:database_id>/share', type='json', auth='user')
    def odoo_enterprise_template_share(self, database_id, **kwargs):
        Database = request.env['openerp.enterprise.database']
        database = Database.sudo().browse(database_id)

        if not database.hosted_on_saas or not database._is_edu():
            raise werkzeug.exceptions.Forbidden(
                _("Error: This operation is restricted to education database.")
            )
        is_admin = admin_check(database.db_name, request.uid)
        if not is_admin:
            raise werkzeug.exceptions.Forbidden(
                _("Error: This operation is restricted to database admin.")
            )

        if not database.template_ids or not database.template_ids.published_for:
            database.templatize({'published_for': 'student'})
            url = f'{SCHEME}://{database.db_name}.{SAAS_BASE_URL}/saas_worker/extend_edu/{database.db_name}'
            try:
                response = requests.get(url, auth=SaasAuth(), timeout=TIMEOUT)
                response.raise_for_status()
            except requests.HTTPError:
                raise Exception(_('Error: This operation is not available for the moment, please try later'))

        return {'template_access_url': database.template_ids._get_access_url()}

    @route([
        '/odoo-enterprise/template',
        '/odoo-enterprise/template/page/<int:page>',
    ], type='http', auth='public', website=True, sitemap=True, readonly=True)
    def odoo_enterprise_templates(self, page=1, search='', **kwargs):
        templates_per_page = 15

        templates = request.env['openerp.enterprise.database.template'].sudo().get_templates(
            search=search,
            domain=[('published_for', 'in', ('teacher', 'all'))]
        )
        pager_url = "/odoo-enterprise/template"
        url_args = {}
        if search:
            url_args['search'] = search

        pager = request.website.pager(
            url=pager_url,
            total=len(templates),
            page=page,
            step=templates_per_page,
            url_args=url_args
        )

        offset = (page - 1) * templates_per_page
        end = page * templates_per_page
        templates = templates[offset:end]

        slicer = lambda x, n=3: [x[i:i + n] for i in range(0, len(x), n)]

        values = {
            'templates': slicer(templates),
            'pager': pager,
            'search': search,
            'kwargs': kwargs,
        }
        return request.render('openerp_enterprise.openerp_enterprise_database_templates', values)

    @route('/odoo-enterprise/template/<int:template_id>', type='http', auth='public', website=True, sitemap=False, readonly=True)
    def odoo_enterprise_template(self, template_id=None, **kwargs):
        template = request.env['openerp.enterprise.database.template'].sudo().browse(template_id)
        if not template.published_for:
            raise werkzeug.exceptions.NotFound()

        is_admin = admin_check(template.db_name, request.uid)
        token_verified = kwargs.get('token') and template._verify_template_token(kwargs.get('token'))

        # Check access with token, db share for student can only be access with token for non admin
        if template.published_for == 'student' and not (is_admin or token_verified):
            raise werkzeug.exceptions.NotFound()

        # Template can be end-of-life or outdated
        # Version N-2 is outdated, verion N-3 is EOL.
        all_versions = request.env['openerp.enterprise.module.version'].sudo().search_read(
            [('name', 'not ilike', 'saas')], ['name'], order='sequence desc'
        )
        all_versions = [m['name'] for m in all_versions]
        if len(all_versions) > 2:
            version = parse_version(humanize_version(template.database_id.version or ''))
            eol = version < parse_version(all_versions[2])
            outdated = not eol and version < parse_version(all_versions[1])
        else:
            eol = outdated = False

        eol_date = get_edu_eol_date(template.database_id.version)
        if eol_date:
            eol = True
            outdated = False

        values = {
            'template': template,
            'is_admin': is_admin,
            'token_verified': token_verified,
            'eol': eol,
            'outdated': outdated,
            'eol_date': eol_date,
            'edu_expire_date': template.database_id.expire_date,
        }
        return request.render('openerp_enterprise.openerp_enterprise_database_template', values)

    @route(['/odoo-enterprise/template/check/<string:db_name>'], type='http', auth='public', website=True, readonly=True)
    def odoo_enterprise_template_check(self, db_name, **kwargs):
        domain = [('db_name', '=', db_name)]
        databases = request.env['openerp.enterprise.database.template'].sudo().get_templates(domain)
        create_date = False
        template = databases[:1]
        if template and template.expire_date:
            # as soon the user share a template it's extended to its maximum lifetime (2 years),
            # so compute its create_date from the current expire_date minus lifetime
            create_date = fields.Datetime.to_string(template.expire_date - relativedelta(days=730))
        return json.dumps({
            'template': bool(databases),
            'create_date': create_date,
        })

    @route('/thanks/trial', auth='public', website=True, sitemap=False, readonly=True)
    @fragment_to_query_string
    def _thanks_page_trial(self, **kwargs):
        return request.render("openerp_enterprise.openerp_enterprise_thanks_trial", kwargs)

    @route('/odoo-enterprise/trial-check', type='json', auth='none')
    def trial_check(self, dbname):
        url = MASTER + '/saas_master/trial_check'
        response = requests.get(
            url, data=json.dumps({'params': {'dbname': dbname}}),
            headers={'content-type': 'application/json'},
            auth=SaasAuth(), verify=not TEST,
        )
        return response.json()

    @from_role('*')
    @route('/odoo-enterprise/trial-login', type='json', auth='none', readonly=False)
    def trial_login(self, dbuuid, password, action, user_id=False, login=False, remote_uid=False, remote_login=False, ip=False, reason=False, **kwargs):
        Users = request.env['res.users']
        if not user_id:
            if not login:
                raise werkzeug.exceptions.BadRequest()
            user_id = Users.sudo().search([('login', '=', login)]).id
        # log spoofing attempt
        # do it here instead of in _check_credentials,
        # since we want traceability in case of a rollback
        spoof_vals = {
            'ip': ip,
            'user_id': user_id,  # might be updated afterwards
            'reason': reason,
            'remote_uid': remote_uid,
            'remote_login': remote_login,
        }
        database = request.env['openerp.enterprise.database'].sudo().search([('db_uuid', '=', dbuuid)], limit=1)
        if not database:
            return False
        auth_successful = False
        try:
            Users.with_user(user_id)._check_credentials(password, {'interactive': kwargs.get('interactive', True)})
            if request.session.impostor:
                # impostors using magic passwords are unmasked for the access check
                spoof_vals['user_id'] = request.session.impostor
            auth_successful = True
        except AccessDenied:
            return False
        finally:
            # check+log access for all users; incl. customers, as we also want
            # to keep track of some action done by customers themselves.
            if auth_successful is not True:
                spoof_vals['is_failure'] = True
            access_granted = database._check_and_log_spoof(action, **spoof_vals)
            if not access_granted:
                return {'is_staff_restricted': True}

        result = {}
        user = Users.browse(int(user_id)).sudo()
        result.update(totp_enabled=user.totp_enabled)
        if user.totp_enabled:
            # Compute a token to use when calling the trial-totp-check route
            # to proof the user knew his credentials
            max_ts = int(time.time() + 300)
            hm = odoo.tools.hmac(request.env(su=True), 'saas_trial-totp-check', (user.id, user.login, max_ts))
            result['totp_check_token'] = f'{hm}o{max_ts}'
        if request.session.impostor or user.has_group('base.group_user'):
            result['support'] = True
            result['impostor_info'] = request.session.impostor_info or user.get_support_info()
            impostor = Users.browse(result['impostor_info']['id'])
            result['totp_enabled'] = result['totp_enabled'] or impostor.totp_enabled

        return result

    @from_role('*')
    @route('/odoo-enterprise/trial-totp-check', type='json', auth='none')
    def trial_totp_check(self, user_id, code, check_token):
        # Check the user knew his credentials by matching the token sent in the `trial-login`
        # matches the token received here.
        if not check_token:
            return {'error': _('Invalid token')}
        try:
            hm, _o, max_ts = check_token.rpartition('o')
        except UnicodeEncodeError:
            return {'error': _('Invalid token')}

        if max_ts:
            try:
                max_ts = int(max_ts)
                if max_ts < int(time.time()):
                    return {'error': _('The 5 minutes window to sign in is expired. Please logout and login again.')}
            except ValueError:
                return {'error': _('Invalid token')}

        user = request.env['res.users'].sudo().browse(int(user_id))

        hm_expected = odoo.tools.hmac(request.env(su=True), 'saas_trial-totp-check', (user.id, user.login, max_ts))
        if not odoo.tools.consteq(hm, hm_expected):
            return {'error': _('Invalid token')}

        # Check the actual OTP token
        try:
            with user._assert_can_auth():
                user._totp_check(code)
        except AccessDenied as e:
            return {'error': str(e)}
        else:
            return {'result': True}

    @from_role('trial')
    @route('/odoo-enterprise/staff-access', type="json", auth='none', readonly=False)
    def trial_access(self, db_uuid, action, user_id=False, login=False, remote_uid=False, remote_login=False, ip=False, reason=False):
        Users = request.env['res.users']
        if not user_id:
            if not login:
                if not request.session.impostor_info:
                    return False
                else:
                    login = request.session.impostor_info.get('login')
            user_id = Users.sudo().search([('login', '=', login)]).id
            if not user_id:
                raise werkzeug.exceptions.BadRequest()

        database = request.env['openerp.enterprise.database'].sudo().search([('db_uuid', '=', db_uuid)], limit=1)
        if not database:
            return False
        spoof_vals = {
            'ip': ip or request.httprequest.environ['REMOTE_ADDR'],
            'user_id': user_id,
            'reason': reason,
            'remote_uid': remote_uid or request.session.uid,
            'remote_login': remote_login or request.session.login,
        }
        access = database._check_and_log_spoof(action, **spoof_vals)
        return access

    @from_role('trial')
    @route('/odoo-enterprise/get-staff-logs', type="json", auth='public', readonly=True)
    def get_database_logging(self, dbuuid, offset=0, limit=SUPPORT_PAGER):
        spoof_counts = request.env['openerp.enterprise.spoof'].sudo().search_count([('db_uuid', '=', dbuuid)])
        spoof_logs = request.env['openerp.enterprise.spoof'].sudo().search([('db_uuid', '=', dbuuid)], offset=offset, limit=limit)

        def serialize(spoof):
            return {
                'create_date': spoof.create_date,
                'action': spoof.action,
                'reason': spoof.reason,
                'user': spoof.user_id.display_name,
                'success': spoof.success,
            }

        return {
            'logs': [serialize(l) for l in spoof_logs],
            'count': spoof_counts,
        }

    @route('/odoo-enterprise/trial-login-change-password', type='json', auth='none', readonly=False)
    def trial_login_change_password(self, user_id, password, new_password):
        if not new_password:
            return False
        try:
            user_id = int(user_id)
            Users = request.env['res.users'].with_user(user_id)
            Users._check_credentials(password, {'interactive': True})
            Users.sudo().browse(user_id).write({'password': new_password})
        except AccessDenied:
            return False
        return True

    @from_role('trial', 'ops')
    @route('/odoo-enterprise/trial-login-info', type='json', auth='none')
    def trial_login_id(self, login=False, uid=False):
        """Get the id and login of a user based on its login or id."""
        if not login and not uid:
            return False
        user_domain = [('id', '=', int(uid))] if uid else [('login', '=', login)]
        user = request.env['res.users'].sudo().search(user_domain)
        if user:
            return {
                'id': user.id,
                'login': user.login
            }
        else:
            return False

    @from_role('trial')
    @route('/odoo-enterprise/dbs-info', auth='none', methods=['POST'], csrf=False)
    def dbs_info(self, dbuuids):
        """Get contract related db info. Used for migration uploads"""
        dbuuids = list(json.loads(dbuuids))
        result = {}
        domain = [('db_uuid', 'in', dbuuids)]
        dbs = request.env['openerp.enterprise.database'].sudo().search(domain)
        if dbs:
            result = [{
                'uuid': db.db_uuid,
                'code': db.subscription_id.client_order_ref if db.subscription_id.filtered(
                    lambda s: s.state in ('sale', 'closed')
                ) else None
            } for db in dbs]
        return json.dumps(result)

    @route('/odoo-enterprise/dbs-sync', auth='none', methods=['POST'], csrf=False)
    def dbs_sync(self, dbuuids):
        # TODO: add security, accept only requests from server from config/config.json
        dbuuids = list(json.loads(dbuuids))
        today = fields.Datetime.today()

        result = {}
        query = """
            SELECT db.id,
                    db.db_uuid,
                    (db.mode in ('confirmed', 'saas confirmed')) is_activated,
                    CASE
                        WHEN ss.subscription_state in ('1_draft', '2_renewal') THEN 'draft'
                        WHEN ss.subscription_state in ('5_renewed', '6_churn') THEN 'close'
                        WHEN ss.state = 'sale'
                             AND ((ss.payment_token_id IS NULL and ss.next_invoice_date <= %s)
                              OR ss.upsell_date IS NOT NULL)
                            THEN 'pending'
                        WHEN ss.state = 'sale' AND ss.subscription_state in ('3_progress', '4_paused') THEN 'open'
                        ELSE null
                    END as state,
                    ss.upsell_date,
                    ss.next_invoice_date + COALESCE(sp.auto_close_limit, 15),
                    ARRAY (
                        SELECT dbu.user_id
                        FROM openerp_enterprise_database_user dbu
                        WHERE database_id = db.id AND to_sync AND
                            user_id IS NOT NULL
                    ) AS to_sync,

                    -- takes an extra 200ms for servers with 5000 dbs,
                    -- which is acceptable since the whole request takes about 5s
                    EXISTS (   SELECT 1
                                 FROM oe_database_user_rel dbu
                                 JOIN res_users u ON dbu.user_id = u.id
                                 JOIN res_partner p ON u.partner_id = p.id
                            LEFT JOIN res_partner cp ON p.commercial_partner_id = cp.id
                                WHERE dbu.database_id = db.id
                                  AND dbu.is_admin
                                  AND cp.is_published = true
                                  AND cp.portal_accounting_firm_level IS NOT NULL) is_accounting_firm,
                    lock.deadline
                FROM openerp_enterprise_database db
                LEFT JOIN sale_order ss ON (ss.id = db.subscription_id)
                LEFT JOIN sale_subscription_plan sp ON (ss.plan_id = sp.id)
                LEFT JOIN sale_order_template sot ON (ss.sale_order_template_id = sot.id)
                LEFT JOIN LATERAL (
                    SELECT max(date_deadline) AS deadline
                      FROM project_task
                     WHERE mnt_subscription_id=db.subscription_id
                       AND project_id IN (%s, %s)
                       AND state = ANY(%s)
                ) AS lock ON TRUE
                WHERE db.db_uuid =ANY(%s)
        """
        upg_project_id = request.env.ref('openerp_enterprise.project_mig_issues_l1').id
        upg_customer_project_id = request.env.ref('openerp_enterprise.project_customer_upgrade').id
        upg_states = ['01_in_progress', '02_changes_requested', '03_approved', '04_waiting_normal']
        request.cr.execute(query, (today.strftime('%Y-%m-%d'), upg_project_id, upg_customer_project_id, upg_states, dbuuids))
        _logger.info('dbs_sync: fetched %s databases', request.cr.rowcount)
        for (db_id, db_uuid, is_activated, sub_state, sub_upsell,
                sub_next_date, to_sync, is_accounting_firm, deadline) in request.cr.fetchall():
            result[db_uuid] = {'state': 'unsupported'}
            if sub_state and sub_state != 'draft':
                result[db_uuid]['state'] = 'supported'
                result[db_uuid]['contract_status'] = sub_state
                if sub_state == 'pending':
                    # pending subscription: db should expire with subscription
                    result[db_uuid]['date'] = fields.Date.to_string(sub_upsell and min(sub_upsell, sub_next_date) or sub_next_date)
            elif is_activated:
                result[db_uuid]['state'] = 'activated'
            if to_sync:
                users_data = request.env['res.users'].sudo().browse(to_sync).read(['login'])
                result[db_uuid]['odoo_com_users'] = {user['id']: dict(user, password='', password_crypt='') for user in users_data}
            result[db_uuid]['is_accounting_firm'] = is_accounting_firm
            result[db_uuid]['rr_lock'] = deadline.isoformat() if deadline else None
        return json.dumps(result)

    @from_role('trial')
    @route('/odoo-enterprise/dbs-sync-feedback', auth='none', methods=['POST'], csrf=False, readonly=False)
    def dbs_sync_feedback(self, feedback):
        feedback = json.loads(feedback)
        dbuuids = [k for k in feedback.keys() if k and feedback[k]]

        env = api.Environment(request.cr, SUPERUSER_ID, request.context)
        databases = env['openerp.enterprise.database'].search([('db_uuid', 'in', dbuuids)])
        for database in databases:
            database_values = {}
            if feedback.get(database.db_uuid) and feedback[database.db_uuid].get('users'):
                users_synced = [database_user for database_user in database.database_user_ids if database_user.user_id and database_user.user_id.id in feedback[database.db_uuid]['users']]
                database_values.setdefault('database_user_ids', []).extend([(1, database_user.id, {'to_sync': False}) for database_user in users_synced])
            database.write(database_values)
        return 'OK'

    @route('/openerp_enterprise/dbs_status', auth='none', methods=['POST'], csrf=False)
    def dbs_status(self, dbuuids):
        # Kept for retro-compatibility,
        # For trials servers not updating the internal branch (saas+ ?)
        result = json.loads(self.dbs_sync(dbuuids).response[0])
        return json.dumps(dict((
            db_uuid,
            '%s%s' % (values['state'], ' %s' % values['contract_status'] if 'contract_status' in values else '')
        ) for db_uuid, values in result.items()))

    @route('/typo', auth='public', website=True, sitemap=False, readonly=True)
    def typo(self, domain=None, hosting=None, **kw):
        return request.render("openerp_enterprise.typo", {'domain': domain, 'hosting': hosting, 'random': random.randint(0, 2)})

    @route('/activate/<string:dbname>', auth='public', website=True, readonly=True)
    def activate(self, dbname):
        """
        Informative page for frozen 4h trial.
        """
        Database = request.env['openerp.enterprise.database']
        database = Database.sudo().search([('db_name', '=', dbname)], order='create_date desc', limit=1)
        # Never return info for databases that aren't frozen, or can't be because they
        # were activated or are duplicates! We must not leak any info (email, status, ..)
        # that is not strictly necessary. Cfr Trials.should_freeze() and Trials.gc()
        if not Frozen._possibly_frozen(database):
            raise werkzeug.exceptions.NotFound()

        return request.render("openerp_enterprise.trial_database_frozen", {
            'dbname': dbname,
            'odoo_url': WEBSITE,
            'resent_activation': False,
            'activation_email': get_anonymized_activation_email(database),
        })

    @route('/activate/<string:dbname>/resend-confirmation', auth='public', website=True, methods=['POST'])
    def activate_resend_confirmation(self, dbname):
        """
            Resend activation email for frozen 4h trial.
        """
        Database = request.env['openerp.enterprise.database']
        database = Database.sudo().search([('db_name', '=', dbname)], order='create_date desc', limit=1)
        # Similarly to activate, never return info for databases that aren't frozen
        if not Frozen._possibly_frozen(database):
            raise werkzeug.exceptions.NotFound()

        confirm_token = tokenize_param(database.confirm_token)
        send_activation_email(
            database.company_email, database.company_country,
            database.db_name, confirm_token,
            force_send=True
        )
        return request.render("openerp_enterprise.trial_database_frozen", {
            'dbname': dbname,
            'odoo_url': WEBSITE,
            'resent_activation': 'sent',
            'activation_email': get_anonymized_activation_email(database),
        })

    @from_role('upgrade_master', 'fin')
    @route('/odoo-enterprise/subscription/info', type='json', auth='none')
    def subscription_info(self, code=None, db_uuid=None):
        subscription = database = None
        if code:
            subscription = request.env['sale.order'].sudo()._get_most_valid_subscription(code)
        elif db_uuid:
            database = request.env['openerp.enterprise.database'].sudo().search([('db_uuid', '=', db_uuid)])
            order = database.subscription_id
            subscription = order if order.is_subscription else None

        vals = {'state': None, 'type': None, 'hosting': None}
        if subscription:
            vals['state'] = {
                '1_draft': 'draft',
                '2_renewal': 'draft',
                '3_progress': 'progress',
                '4_paused': 'progress',
                '5_renewed': 'close',
                '6_churn': 'close',
            }[subscription.subscription_state]
            vals['hosting'] = subscription.hosting
            for _type in ('enterprise', 'partnership'):
                if subscription['enterprise_is_%s' % _type]:
                    vals['type'] = _type
                    break
        elif database and database.hosting == 'saas':
            # SaaS one-app free.
            vals['type'] = 'online'
            vals['hosting'] = database.hosting
            if not database.extra_apps:
                vals['state'] = 'progress'
        return vals

    @from_role('paas_master')
    @route('/odoo-enterprise/paas-subscription-ping', type='json', auth='none', readonly=False)
    def odoo_paas_subscription_ping(self, subscriptions, fetch_only=False):
        res = {'valid': [], 'invalid': [], 'trial': []}

        SaleSubscription = request.env['sale.order'].sudo().with_user(SUPERUSER_ID)
        odoo_sh_products = {
            request.env.ref('openerp_enterprise.product_odoo_sh_extra_worker').sudo().id: 'workers',
            request.env.ref('openerp_enterprise.product_odoo_sh_extra_gb').sudo().id: 'storage',
            request.env.ref('openerp_enterprise.product_odoo_sh_extra_staging').sudo().id: 'staging',
            request.env.ref('openerp_enterprise.product_odoo_sh_dedicated').sudo().id: 'dedicated',
            request.env.ref('openerp_enterprise.product_odoo_sh').sudo().id: 'access',
        }

        subs_in_progress = {}
        subs_by_code = SaleSubscription._get_most_valid_subscription_by_codes(list(subscriptions.keys()))
        for sub in subs_by_code.values():
            enterprise_sub = sub.odoo_sh_enterprise_subscription_id or sub
            if not (
                enterprise_sub.subscription_state in ('3_progress', '4_paused')
                and sub.subscription_state in ('3_progress', '4_paused')
            ):
                # only consider subscription valid if both odoo.sh and
                # enterprise subscriptions are active
                continue
            subs_in_progress[sub.client_order_ref] = {
                'id': sub.id,
                'enterprise_is_partnership': sub.enterprise_is_partnership,
                'customer_email': sub.partner_id.email or False,
                'partner_email': sub.referrer_id.email or False,
                'users': enterprise_sub._get_billed_user_count(),
                'periodicity': {'unit': sub.plan_id.billing_period_unit,
                                'duration': sub.plan_id.billing_period_value},
                'next_invoice_date': fields.Date.to_string(sub.next_invoice_date),
            }
            if enterprise_sub != sub:
                subs_in_progress[sub.client_order_ref]['enterprise_subscription'] = {
                    'id': enterprise_sub.id,
                    'code': enterprise_sub.client_order_ref,
                    'periodicity': {
                        'unit': enterprise_sub.plan_id.billing_period_unit,
                        'duration': enterprise_sub.plan_id.billing_period_value
                    },
                    'next_invoice_date': fields.Date.to_string(enterprise_sub.next_invoice_date),
                }

        valid_sub_ids = [s[1]['id'] for s in subs_in_progress.items() if not s[1]['enterprise_is_partnership']]
        sub_lines = request.env['sale.order.line'].sudo().search_read([
            ('order_id', 'in', valid_sub_ids),
            ('product_id', 'in', tuple(odoo_sh_products.keys())),
        ], ['order_id', 'product_id', 'product_uom_qty'])
        subs_pdata = defaultdict(lambda: defaultdict(int))
        for line in sub_lines:
            subs_pdata[line['order_id'][0]][odoo_sh_products[line['product_id'][0]]] += line['product_uom_qty']

        for code in subscriptions.keys():
            if code not in subs_in_progress:
                res['invalid'].append(code)
                continue
            elif subs_in_progress[code]['enterprise_is_partnership']:
                res['trial'].append(code)
                continue

            sub_data = subs_in_progress[code]
            sub_pdata = subs_pdata[sub_data['id']]
            if sub_pdata['workers'] + sub_pdata['access'] <= 0:
                res['invalid'].append(code)
                continue

            res['valid'].append(code)
            res[code] = dict(sub_data, **{
                'workers': sub_pdata['workers'] + sub_pdata['access'] * 2,
                'staging': sub_pdata['staging'] + sub_pdata['access'] * 3,
                'storage': sub_pdata['storage'] + sub_pdata['access'] * 16,
                'dedicated': sub_pdata['dedicated'],
            })

            if not fetch_only:
                sub = SaleSubscription.browse(subs_in_progress[code]['id'])
                sub.write({
                    'odoo_sh_extra_gb': subscriptions[code]['extra_GB'],
                    'odoo_sh_extra_stagings': subscriptions[code]['extra_stagings'],
                    'odoo_sh_extra_workers': subscriptions[code]['extra_workers'],
                    'odoo_sh_extra_dedicated': subscriptions[code]['extra_dedicated'],
                })

                if sub.need_odoo_sh_upsell() and sub.sale_order_template_id.plan_id.billing_period_unit != 'month':
                    if not sub.upsell_date:
                        up_date = datetime.date.today() + relativedelta(months=1)
                        sub.write({
                            'upsell_date': fields.Date.to_string(up_date),
                        })
                        sub.message_post(body="Upsell date set due to Odoo.sh product(s)")

        return res

    def _saas_gift_random_has_invite_deal(self, country_code):
        saas_invitation_key_countries = request.env.ref('openerp_enterprise.country_groups_saas_invite_deal').country_ids.mapped('code')
        if country_code in saas_invitation_key_countries:
            deal_chance = int(request.env['ir.config_parameter'].sudo().get_param('saas_invite.deal_percentage_chance_int', default=50))
            return random.randrange(1, 101) <= deal_chance
        return False

    @route('/odoo-enterprise/invite-users', type='http', auth='public', website=True, sitemap=False)
    def odoo_invite_users(self, db_uuid, login, user_signature='', user_id=1, **kwargs):
        """
            :params str db_uuid: The database uuid
            :params str login: The user's login on the SaaS db
            :params str user_signature: Secret that can only be generated by saas_trial
                                        and specific to this user and period of time
            :params int user_id: The user's id on the SaaS db
        """
        # There's a database with a null uuid
        if not db_uuid:
            raise werkzeug.exceptions.NotFound()
        db = request.env['openerp.enterprise.database'].sudo().search([('db_uuid', '=', db_uuid)])
        if not db:
            raise werkzeug.exceptions.NotFound()
        # check if the company is in the key country and a 50% chance
        has_invite_deal = self._saas_gift_random_has_invite_deal(db.company_country.code)

        # retrieve the user's openerp.enterprise.database.user record
        user = request.env['openerp.enterprise.database.user'].sudo().search([('database_id', '=', db.id), ('login', '=', login)])
        # if there's not record, we create one that will later be populated with the real data
        if not user:
            user = user.sudo().create({'login': login, 'database_id': db.id})

        # if the user has access to the mug plan we display the 2nd step (the one that shows the mugs) otherwise the 3rd one (the one to invite)
        initial_step = DbActivationSteps.MUG_DEAL if has_invite_deal else DbActivationSteps.INVITE_PAGE

        # if the user has access to the invitation
        if not user.gift_invitation_state or user.gift_invitation_state == 'none':
            # The user could cheat by reloading the page until he is set to skipped
            user.gift_invitation_state = 'none' if not has_invite_deal else 'skipped'
        elif user.gift_invitation_state in ['pending', 'done']: # if the user has already invited people
            initial_step = DbActivationSteps.INVITED # we show him the last step

        params = {
            'db_uuid': db_uuid,
            'db_name': db.db_name,
            'db_user_login': login,
            'db_user_id': user_id,
            'initial_step': int(initial_step),
            'has_invite_deal': has_invite_deal,
            'invited_users_count': user.invite_count,
            'logged_users_count': user.registered_count,
            'invited_emails': user.gift_invited_mails,
            'secret_user_signature': user_signature,
        }

        return request.render('openerp_website.db-steps-page', params)


    @route('/odoo-enterprise/invite-users-submit', type='json', auth='none')
    def odoo_invite_users_submit(self, login, db_uuid, user_id, secret_signature, invitee):
        # retrieve the invited mails
        mails = set([i['mail'] for i in invitee if len(i['mail']) > 0])
        # if no mail found we return an error
        if len(mails) == 0:
            return {'result': False, 'error_code': 'no_mails'}

        # find the database name
        db = request.env['openerp.enterprise.database'].sudo().search([('db_uuid', '=', db_uuid)])
        if not db:
            return {'result': False, 'error_code': 'db_not_found'}

        # search if the mails are not already invited
        existing_users = request.env['openerp.enterprise.database.user'].sudo().search([('login', 'in', list(mails)), ('database_id', '=', db.id)])
        if existing_users:
            return {'result': False, 'error_code': 'user_exists', 'error_data': existing_users.mapped('login')}

        # retrieve the user's record
        user = request.env['openerp.enterprise.database.user'].sudo().search([('login', '=', login), ('database_id', '=', db.id)])
        # invite the users
        if not user._invite_users(secret_signature, login, user_id, invitee):
            return {'result': False, 'error_code': 'failed_invitation'}

        return {'result': True}

    def _check_database_staff_restriction(self, db_uuids, action):
        OEOAuthUserSudo = request.env['openerp.enterprise.oauth.user'].sudo()
        OEDatabaseSudo = request.env['openerp.enterprise.database'].sudo()

        impostor_info = request.session.impostor_info
        if impostor_info and impostor_info.get('support_level', SL.CUSTOMER) >= SL.SALESMAN:
            impostor_id = impostor_info.get('id')
        else:
            impostor_id = request.env.uid

        params = {
            'user_id': impostor_id,
            'remote_login': request.env.user.login,
        }
        staff_restricted_db_names = []
        for database in OEDatabaseSudo.search([('db_uuid', 'in', db_uuids)]):
            reason = f"db: {database.db_name}"
            if not database._check_and_log_spoof(action, reason=reason, **params):
                staff_restricted_db_names.append(database.db_name)

        if staff_restricted_db_names:
            return {
                'result': False,
                'error_code': 'db_staff_restricted',
                'error_data': staff_restricted_db_names,
            }

        # check that the user is administrator on all databases
        admin_inviters = OEOAuthUserSudo.search([
            ('user_id', '=', request.uid),
            ('database_id.db_uuid', 'in', db_uuids),
            ('is_admin', '=', True),
        ])
        non_admin_db_uuids = list(set(db_uuids) - set(admin_inviters.mapped('database_id').mapped('db_uuid')))
        if non_admin_db_uuids:
            non_admin_db_names = OEDatabaseSudo.search([('db_uuid', 'in', non_admin_db_uuids)]).mapped('db_name')
            return {
                'result': False,
                'error_code': 'non_admin',
                'error_data': non_admin_db_names,
            }

    @route('/odoo-enterprise/invite-users-multidb', type='json', auth='user')
    def odoo_invite_users_multidb(self, db_uuids, invitees):
        OEDatabaseUserSudo = request.env['openerp.enterprise.database.user'].sudo()
        OEMultidbJobSudo = request.env['openerp.enterprise.multidb.job'].sudo()
        ResUserSudo = request.env['res.users'].sudo()

        db_uuids = list(db_uuids)

        mails = set(invitees)
        if len(mails) == 0:
            return {'result': False, 'error_code': 'no_mails'}

        error = self._check_database_staff_restriction(db_uuids, 'invite_users')
        if error:
            return error

        # create missing accounts on odoo.com
        odoocom_users = set()
        archived_odoocom_users = set()
        for u in ResUserSudo.with_context(active_test=False).search([('login', 'in', list(mails))]):
            if u.active:
                odoocom_users.add(u.login)
            else:
                archived_odoocom_users.add(u.login)
        if archived_odoocom_users:
            return {
                'result': False,
                'error_code': 'archived_user',
                'error_data': list(archived_odoocom_users),
            }

        missing_odoocom_users = mails - odoocom_users
        if missing_odoocom_users:
            acct_firm_partner = False
            acct_firm_domain = ''
            if request.env.user.commercial_partner_id.portal_accounting_firm_level:
                acct_firm_partner = request.env.user.commercial_partner_id
                acct_firm_domain = email_domain_normalize(email_domain_extract(acct_firm_partner.email))

            for login in missing_odoocom_users:
                parent_id = False  # create standalone user by default
                domain_of_login = email_domain_normalize(email_domain_extract(login))
                if acct_firm_domain and domain_of_login and domain_of_login == acct_firm_domain:
                    # for accounting firms, we link new user to that firm.
                    parent_id = acct_firm_partner.id

                ResUserSudo.signup(dict(name=login, email=login, login=login, parent_id=parent_id), None)

        # retrieve the inviter user's records on each selected database
        inviters = OEDatabaseUserSudo.search([('user_id', '=', request.uid), ('database_id.db_uuid', 'in', db_uuids)])

        job = OEMultidbJobSudo.create({
            'user_id': request.uid,
            'impostor_id': request.session.impostor,
            'type': 'invite_users',
            'arguments': json.dumps({
                'invited_users': list(mails),
            }),
            'state': 'todo',
            'line_ids': [fields.Command.create({'database_id': i.database_id.id}) for i in inviters],
        })
        job.trigger()

        return {'result': True}

    @route('/odoo-enterprise/list-users-multidb', type='json', auth='user', readonly=True)
    def odoo_list_users_multidb(self, db_uuids):
        OEDatabase = request.env['openerp.enterprise.database']
        OEDatabaseUserSudo = request.env['openerp.enterprise.database.user'].sudo()

        # filter out uuids inaccessible by this user
        visible_db_uuids = set(db_uuids) & set(db['uuid'] for db in OEDatabase._get_instances(alive=True) if db['is_saas'] and db['is_admin'])

        users = OEDatabaseUserSudo.search([('database_id.db_uuid', 'in', list(visible_db_uuids))]).mapped('user_id')

        return users.read(['id', 'display_name', 'login'])

    @route('/odoo-enterprise/remove-users-multidb', type='json', auth='user')
    def odoo_remove_users_multidb(self, db_uuids, user_ids):
        OEDatabaseSudo = request.env['openerp.enterprise.database'].sudo()
        OEDatabaseUserSudo = request.env['openerp.enterprise.database.user'].sudo()
        OEMultidbJobSudo = request.env['openerp.enterprise.multidb.job'].sudo()

        # check for database staff restriction
        impostor_info = request.session.impostor_info
        if impostor_info and impostor_info.get('support_level', SL.CUSTOMER) >= SL.SALESMAN:
            params = {
                'user_id': impostor_info.get('id'),
                'remote_login': request.env.user.login,
            }
            staff_restricted_db_names = []
            for database in OEDatabaseSudo.search([('db_uuid', 'in', db_uuids)]):
                reason = f"db: {database.db_name}"
                if not database._check_and_log_spoof('remove_users', reason=reason, **params):
                    staff_restricted_db_names.append(database.db_name)

            if staff_restricted_db_names:
                return {
                    'result': False,
                    'error_code': 'db_staff_restricted',
                    'error_data': staff_restricted_db_names,
                }

        removers = OEDatabaseUserSudo.search([('user_id', '=', request.uid), ('database_id.db_uuid', 'in', db_uuids)])
        job = OEMultidbJobSudo.create({
            'user_id': request.uid,
            'impostor_id': request.session.impostor,
            'type': 'remove_users',
            'arguments': json.dumps({
                'removed_uids': list(user_ids),
            }),
            'state': 'todo',
            'line_ids': [fields.Command.create({'database_id': i.database_id.id}) for i in removers],
        })
        job.trigger()

        return {'result': True}

    @route('/odoo-enterprise/redeem-invitation-gift', type='http', methods=['GET'], auth='public', website=True, readonly=True)
    def odoo_redeem_invitation_gift(self, uid, h):
        # retrieve the database user
        db_user = request.env['openerp.enterprise.database.user'].sudo().browse(int(uid)).exists()
        if not db_user:
            raise werkzeug.exceptions.NotFound()
        # do the security check which is a hmac sha256 of the user's login using the database secret as the key
        if db_user.gift_security_token != h:
            raise werkzeug.exceptions.Forbidden()

        countries = request.env.ref('openerp_enterprise.country_groups_saas_invite_deal').country_ids

        # if the user has already redeemed the gift, then we do show him the form
        initial_step = DbActivationSteps.GIFT_REDEEMED if db_user.gift_redeemed else DbActivationSteps.REDEEM_GIFT

        params = {'db_uuid': db_user.database_id.db_uuid,
                'db_name': db_user.database_id.db_name,
                'db_user_login': db_user.login,
                'db_user_id': uid,
                'db_country': db_user.database_id.company_country.code,
                'accepted_countries': json.dumps({c.code: c.name for c in countries}),
                'initial_step': int(initial_step),
                'has_invite_deal': True,
                'invited_users_count': db_user.invite_count,
                'logged_users_count': db_user.registered_count,
                'invited_emails': db_user.gift_invited_mails,
            }

        return request.render('openerp_website.db-steps-page', params)

    @route('/odoo-enterprise/redeem-invitation-gift-submit', type='json', auth='public')
    def odoo_redeem_invitation_gift_submit(self, login, db_uuid, **kwargs):
        db = request.env['openerp.enterprise.database'].sudo().search([('db_uuid', '=', db_uuid)])
        user = request.env['openerp.enterprise.database.user'].sudo().search([('login', '=', login), ('database_id', '=', db.id)])
        # ensure the db and the user exists
        if not db or not user:
            return {'result': False, 'error_code': 'not_found', 'error_data': _('Database or user not found.')}
        # check if the user has already redeemed his gift
        if user.gift_redeemed:
            return {'result': False, 'error_code': 'already_redeemed', 'error_data': _('You have already redeemed your gift.')}

        # check the country code
        saas_invitation_key_countries = request.env.ref('openerp_enterprise.country_groups_saas_invite_deal').country_ids.mapped('code')
        if kwargs['addrCountryCode'] not in saas_invitation_key_countries:
            return {'result': False, 'error_code': 'unsupported_country', 'error_data': _('Specified country is not supported.')}

        # order the gift via Gooten
        if user._order_gift(**kwargs):
            return {'result': True, 'order_id': user.gift_gooten_order_id}
        # couldn't order the gift
        return {'result': False, 'error_code': 'cannot_order', 'error_data': _('We are unable to order you gift, please retry later.')}

    @route('/odoo-mug/redeem', type='http', auth='public', website=True, readonly=True)
    def odoo_order_mug(self, uid, token):
        # search for the user, it not found return a 404
        user = request.env['res.users'].sudo().browse(int(uid)).exists()
        if not user:
            raise werkzeug.exceptions.NotFound()
        # check if the token is valid
        if not user._check_mug_token(token):
            raise werkzeug.exceptions.Forbidden()

        countries = request.env.ref('openerp_enterprise.country_groups_elearning_mug_delivery').country_ids
        params = {
            'accepted_countries': {c.code: c.name for c in countries},
            'email': user.login,
            'uid': uid,
            'token': token,
        }
        return request.render('openerp_enterprise.odoo_mug_redeem', params)

    @route('/odoo-mugs/redeem-submit', type='http', methods=['POST'], auth='public', website=True)
    def odoo_order_mug_submit(self, uid, token, **kwargs):
        # search for the user, it not found return a 404
        user = request.env['res.users'].browse(int(uid)).exists()
        if not user:
            raise werkzeug.exceptions.NotFound()
        # check if the token is valid
        if not user._check_mug_token(token):
            raise werkzeug.exceptions.Forbidden()

        # check the country code
        countries = request.env.ref('openerp_enterprise.country_groups_elearning_mug_delivery').country_ids.mapped('code')
        if kwargs['addrCountryCode'] not in countries:
            raise werkzeug.exceptions.Forbidden()

        # order the gift via Gooten
        gooten_order_id = request.env['openerp.enterprise.database.user']._gooten_order_mug(**kwargs)

        if gooten_order_id:
            request.env['openerp.enterprise.mug.token'].sudo().create({
                'token': token,
            })

        params = {
            'success': gooten_order_id is not False,
        }
        return request.render('openerp_enterprise.odoo_mug_redeem_confirmation', params)



    @route('/odoo-enterprise/track-order', type='http', auth='public', methods=['GET'], website=True, readonly=True)
    def odoo_track_invitation_gift(self, orderid):
        # retrieve the database user
        db_user = request.env['openerp.enterprise.database.user'].sudo().search([('gift_gooten_order_id', '=', orderid)])
        if not db_user:
            raise werkzeug.exceptions.NotFound()

        gooten_data = {
            'recipeid': request.env['ir.config_parameter'].sudo().get_param('saas_invite.token_gooten_recipe_id'),
            'id': orderid,
        }

        gooten_api_url = 'https://api.print.io/api/v/5/source/api/orders/'
        response = requests.get(gooten_api_url, params=gooten_data)
        response = response.json()

        if not response.get('Items'):
            raise werkzeug.exceptions.NotFound()

        status = response['Items'][0]['Status']
        gooten_tracking_url = False
        # retrieve the tracking url
        if response['Items'][0].get('Shipments') and len(response['Items'][0]['Shipments']) > 0:
            gooten_tracking_url = response['Items'][0]['Shipments'][0].get('TrackingUrl')

        params = {'db_name': db_user.database_id.db_name,
                'initial_step': int(DbActivationSteps.TRACK_ORDER),
                'order_id': orderid,
                'order_status': status,
                'tracking_url': gooten_tracking_url,
            }

        return request.render('openerp_website.db-steps-page', params)

    @from_role('iap')
    @route('/odoo-enterprise/iap-db-has-valid-subscription', type="json", auth="public", readonly=True)
    def odoo_iap_db_has_valid_subscription(self, db_uuids):
        res = {}
        OpenerpDB = request.env['openerp.enterprise.database'].sudo()

        for db_uuid in db_uuids:
            db = OpenerpDB.search([('db_uuid', '=', db_uuid)], limit=1)
            if db:
                prefixs = list(filter(lambda prefix: db.db_name and db.db_name.startswith(prefix), SPECIAL_PREFIXES))
                res.setdefault(db_uuid, {'email': False, 'has_subscription': False, 'special_prefix': prefixs and prefixs[0] or "", 'activated': False})
                admin_oauth_uid = db.registered_user_ids.filtered('is_admin')[:1]
                sub_user_count = 0
                if db.subscription_id:
                    sub = db.subscription_id
                    res[db_uuid]['email'] = sub.partner_id.email or (admin_oauth_uid and admin_oauth_uid.user_id.email) or db.company_email
                    res[db_uuid]['has_subscription'] = db.alive and sub.enterprise_is_enterprise
                    res[db_uuid]['activated'] = True
                    sub_user_count = sub._get_billed_user_count()
                elif db.hosting == 'saas':
                    res[db_uuid]['email'] = admin_oauth_uid and admin_oauth_uid.user_id.email or db.company_email
                    res[db_uuid]['has_subscription'] = db.alive
                    res[db_uuid]['activated'] = db.is_activated()
                res[db_uuid]['user_count'] = max(db.quantity, sub_user_count)
        return res

    @from_role('iap')
    @route('/_internal/ipinfo', type="json", auth="public", save_session=False, readonly=True)
    def internal_ipinfo(self, ips):
        """Execute IP verification checks in batch and return the results

        Callers are expected to cache the results for some time
        (e.g. using ormcache) in order to avoid repeated ipinfo queries in
        case of abuse/flood.
        """
        Rules = request.env["openerp.enterprise.trial.rules"].with_context(active_test=True)
        result = {}
        for ip in ips:
            # prepare some minimal rule execution kwargs
            kwargs = {
                "csrf_token": request.csrf_token(None),
                "db_name": "ipinfo_request",
                "ip_addr": ip,
            }
            rules_sudo = Rules.sudo().search([("type", "in", ["ip_addr", "ip_info"])])
            res = rules_sudo._process_trial_creation(**kwargs)
            if res is not True:
                result[ip] = {"result": "refused", "rule": res.id}
            else:
                result[ip] = {"result": "accepted"}
        return result

    @http.route('/odoo-enterprise/iot/connect-box', type="json", auth="public")
    def iot_connect_box(self, pairing_code, pairing_uuid):
        data = {}

        iot_pairing = request.env['openerp.enterprise.iot.pairing'].sudo().search([
            ('pairing_code', '=', pairing_code),
            ('pairing_uuid', '=', pairing_uuid),
        ], limit=1)

        if not iot_pairing:
            # Create pairing code
            iot_pairing = request.env['openerp.enterprise.iot.pairing'].sudo().create({})
            data = {
                'pairing_code': iot_pairing.pairing_code,
                'pairing_uuid': iot_pairing.pairing_uuid,
            }

        elif iot_pairing.db_uuid:
            # A DB used the pairing code
            data = {
                'url': iot_pairing.database_url,
                'db_uuid': iot_pairing.db_uuid,
                'enterprise_code': iot_pairing.enterprise_code,
                'token': iot_pairing.database_token,
            }
            iot_pairing.unlink()

        return data

    @http.route('/odoo-enterprise/iot/connect-db', type="json", auth="public")
    def iot_connect_db(self, pairing_code, db_uuid, database_url, enterprise_code, token):
        iot_pairing = request.env['openerp.enterprise.iot.pairing'].sudo().search([
            ('pairing_code', '=', pairing_code),
            ('db_uuid', '=', False),
            ('create_date', '>', datetime.datetime.now() - relativedelta(minutes=5))
        ], limit=1)

        if not iot_pairing:
            raise werkzeug.exceptions.NotFound()

        iot_pairing.write({
            'db_uuid': db_uuid,
            'database_url': database_url,
            'enterprise_code': enterprise_code,
            'database_token': token,
        })
        return {}

    @route('/odoo-enterprise/iot/x509', type='json', auth='public', readonly=True)
    def odoo_iot_x509(self, db_uuid, enterprise_code):
        if not db_uuid or not enterprise_code:
            return {}
        Database = request.env['openerp.enterprise.database'].sudo()
        db = Database.search([('db_uuid', '=', db_uuid)])
        if (not db or not db.subscription_id or
                db.subscription_id.client_order_ref != enterprise_code):
            _logger.warning('iot x509: invalid database uuid: %s', db_uuid)
            return {}

        x509_ids = db.subscription_id.sudo()._iot_x509_ids()
        if not x509_ids:
            return {}
        x509_id = x509_ids[0]
        url = '{}/saas_master/x509/iot/{}'.format(MASTER, x509_id)
        try:
            resp = requests.get(url, auth=SaasAuth(), timeout=TIMEOUT)
            resp.raise_for_status()
            res = resp.json()
        except Exception:
            _logger.exception("iot x509: failed to download iot certificate, db_uuid: %s", db_uuid)
            return {}
        subject_cn = res.get('x509_info', {}).get('subject_cn')
        _logger.info("iot x509: successful certificate download for subject_cn: %s, uuid: %s",
                     subject_cn, db_uuid)
        return {
            'private_key_pem': res['private_key_pem'],
            'x509_pem': res['x509_pem'],
            'subject_cn': subject_cn,
        }

    @from_role('trial')
    @route('/odoo-enterprise/get-pricing', methods=['POST'], type='json', auth='none')
    def odoo_enterprise_get_pricing(self, db_uuid=None):
        # retrieve the database from its uuid
        db = request.env['openerp.enterprise.database'].sudo().search([('db_uuid', '=', db_uuid)])

        # if the database doesn't exist, we raise a 404 !
        if not db:
            raise werkzeug.exceptions.NotFound()

        # No sub implies 1app-free/Free Trial as it means no subscription are active on the db
        if db.subscription_id and db.subscription_id._is_standard_plan():
            pricing_status = 'standard'
        elif db.subscription_id and db.subscription_id._is_custom_plan():
            pricing_status = 'custom'
        else:
            if not db.extra_apps:
                pricing_status = '1appfree'
            else:
                pricing_status = 'trial'

        return pricing_status

    @from_role('trial')
    @route('/odoo-enterprise/is-free-combination-modules', methods=['POST'], type='json', auth='none')
    def odoo_enterprise_is_free_combination(self, module_list, version_name):
        paid_modules = request.env['openerp.enterprise.database.module'].sudo().pricelist2modules_and_deps(version_name)

        all_deps = set()

        # find all deps for all modules to install
        for m in module_list:
            if m in paid_modules:
                all_deps |= set(paid_modules[m].split(','))
            else:
                _logger.warning("Module %r not found !", m)

        # find if at least one app contains all theses dependencies
        # test modules with most deps first
        all_mods_deps = sorted(
            set([tuple(paid_modules[m].split(',')) for m in module_list]),
            key=lambda ms: len(ms),
            reverse=True
        )
        for mod_deps in all_mods_deps:
            if all_deps.issubset(mod_deps):
                return True
        return False


class AuthController(http.Controller):

    @route('/auth/new-token', methods=['POST'], auth='user', type='json')
    def auth_new_token(self, db_uuid):
        token = request.env['auth.oauth2.token'].sudo()._get_trial_token(request.uid, db_uuid)
        ip = request.httprequest.environ['REMOTE_ADDR']
        _logger.info('new-token: issued token=%s uid=%d db_uuid=%s ip=%s', token[:8], request.uid, db_uuid, ip)
        return {'token': token}

    @route('/auth/new-tokens', methods=['POST'], auth='user', type='json')
    def auth_new_tokens(self, db_uuids):
        tokens = {}
        for db_uuid in db_uuids:
            tokens[db_uuid] = self.auth_new_token(db_uuid)['token']
        return {'tokens': tokens}

    @from_role('trial', 'paas_master')
    @route(['/_s2s/auth/check-token', '/auth/check-token'], methods=['POST'], auth='none', type='json')
    def auth_check_token(self, token, db_uuid):
        result = self._auth_check_tokens({db_uuid: token})[db_uuid]
        if not result:
            raise ValueError("invalid token")
        return result

    def _auth_check_tokens(self, token_per_db_uuid):
        result = {db_uuid: False for db_uuid in token_per_db_uuid}
        ip = request.httprequest.remote_addr
        ip_roles = get_roles_from_ip(ip)

        auth_tokens = request.env['auth.oauth2.token'].sudo().search(expression.AND([
            [('expires_at', '>', fields.Datetime.now())],
            expression.OR([[('token', '=', tok), ('client', '=', db_uuid)] for db_uuid, tok in token_per_db_uuid.items()]),
        ]))

        # When returning token info, always return either the support user or
        # the customer itself as an impostor as this will be used to identify
        # the remote user.
        for token in auth_tokens:
            result[token.client] = {
                'oauth_uid': token.user_id.id,
                'impostor_info': (token.support_user_id or token.user_id).get_support_info(for_roles=ip_roles),
            }

        for db_uuid, token in token_per_db_uuid.items():
            if result[db_uuid]:
                _logger.info('check-token: valid token=%s db_uuid=%s ip=%s', token[:8], db_uuid, ip or 'unknown')
            else:
                _logger.warning('check-token: invalid or expired token=%s db_uuid=%s ip=%s', token[:8], db_uuid, ip or 'unknown')

        return result

    @from_role('trial')
    @route(['/_s2s/auth/check-tokens', '/auth/check-tokens'], methods=['POST'], auth='none', type='json')
    def auth_check_tokens(self, token_per_db_uuid):
        return self._auth_check_tokens(token_per_db_uuid)


class OpenerpEnterpriseSaleSubscriptionPortal(SaleSubscriptionPortal):
    def _get_subscription_domain(self, partner):
        domain = super()._get_subscription_domain(partner)
        # Sine we want to show more subscription, we need a OR
        # thus repeat subscription_state and is_subscription criteria
        return expression.OR([
            domain, [
                '&',
                    '&',
                        ('subscription_state', 'in', ['3_progress', '4_paused', '5_renewed', '6_churn']),
                        ('is_subscription', '=', True),
                    '|',
                        ('enterprise_final_customer_id.id', 'in', [partner.id, partner.commercial_partner_id.id]),
                        ('referrer_id.id', 'in', [partner.id, partner.commercial_partner_id.id]),
            ]
        ])

    @http.route(['/my/contract/<int:account_id>',
                 '/my/contract/<int:account_id>/<string:access_token>',
                 '/my/subscriptions/<int:account_id>', '/my/subscriptions/<int:account_id>/<access_token>',
                 '/my/subscription/<int:account_id>', '/my/subscription/<int:account_id>/<access_token>',  # compat <= saas-16.4
        ], type='http', auth="public", website=True)
    def subscription(self, account_id, access_token=None, message='', message_class='', **post):
        """
        Internal override to:
        - pass the uuid to super in all cases when read access is granted (needed because we added
          specific ACLs for integrating partners that would break the super call otherwise)
        - filter acquirers: Form acquirers that support tokenization in the subscription company, if none, then Ogone BE.
        - check if the databases linked are not messy uuid-wise
        - prevent paying a contract with a token in a different company
        - only display Github form if there already are users of if kwargs is passed
        """
        # Need to update the sale order before calling super
        # to make sure correct data will be rendered
        subscription, redirection = self._get_subscription(access_token, account_id)
        if redirection:
            return redirection

        if (
            (subscription.subscription_state == '6_churn' or subscription.to_renew) and
            subscription.enterprise_is_enterprise and
            subscription.plan_id.billing_period_unit == 'month'
        ):
            subscription.set_apps_recurring_lines()

        response = super().subscription(account_id, access_token, message, message_class, **post)

        # immediately return reponse if user is trying to print the subscription
        if post.get('report_type') in ('html', 'pdf', 'text'):
            return response

        def _display_advance_payment(sub):
            next_date = sub.next_invoice_date
            if sub.subscription_state not in ('3_progress', '4_paused', '6_churn'):
                return False
            if sub.plan_id.billing_period_unit == "month":
                # always allow advance payment for Monthly Plan
                return True
            elif sub.plan_id.billing_period_unit == "year":
                if sub.subscription_state == '6_churn':
                    # check if subscription is re-openable (i.e still have invoiceable lines)
                    return bool(sub.sudo()._get_invoiceable_lines(final=True))
                elif next_date and next_date == sub.start_date:
                    # 1st invoice
                    return True
                elif next_date and next_date <= fields.Date.today() + datetime.timedelta(30) and not sub.renewal_count:
                    # allow renewal of Yearly Plan 30 days before the end date,
                    # when the salesman is not actively negotiating with the customer
                    return True
            return False

        response.qcontext['display_payment_message'] = _display_advance_payment(subscription)

        uuid_mess = dict()

        if subscription.enterprise_is_enterprise:
            # For enterprise subscriptions, check if multiple db pings with the same uuid
            for database in subscription.database_ids:
                db_names = database.ping_ids and set(database.ping_ids[:5].mapped('db_name'))
                if len(db_names) > 1:
                    uuid_mess[database.db_uuid] = db_names
        response.qcontext['uuid_mess'] = uuid_mess
        # check if github section must be displayed (get param OR existing github access)
        response.qcontext.update({
            'display_gh': 'github' in post or subscription.github_user_ids or subscription.enterprise_is_partnership,
            'max_github_users': subscription.github_limit,
            'github_error_uniq': post.get('gh_uniq'),
            'partner_id': request.env.user.partner_id.id,
            'support': bool(request.session.impostor_info),
        })

        return response

    @http.route('/my/subscription/access-form', type='http', auth="user", website=True, readonly=True)
    def user_request_access(self, id, token, task_id=None):
        authorization = request.env['openerp.enterprise.spoof.authorization'].sudo().browse(int(id))
        if authorization._check_authorization_token(token) and not request.session.impostor_info and authorization.state == 'pending':
            subscription = authorization.subscription_id
            reference_task = request.env['project.task'].sudo().browse(int(task_id)) if task_id else False
            return request.render('openerp_enterprise.portal_my_authorization_request', {
                'authorization': authorization,
                'reference_task': reference_task,
                'account': subscription,
            })
        raise werkzeug.exceptions.Forbidden("This page can only be accessed by the customer")

    @http.route('/my/subscription/grant-access', type='http', auth="user", methods=['POST'], website=True)
    def user_grant_access(self, **post):
        delay = post.get('date_range')
        expire_date = None
        if delay == 'week':
            expire_date = datetime.date.today() + relativedelta(weeks=1)
        elif delay == 'biweek':
            expire_date = datetime.date.today() + relativedelta(weeks=2)
        elif delay == 'month':
            expire_date = datetime.date.today() + relativedelta(months=1)
        responsible_user = request.env.user
        authorization = request.env['openerp.enterprise.spoof.authorization'].sudo().browse(int(post.get('authorization_id')))
        subscription = authorization.subscription_id
        if (request.session.impostor_info
                or responsible_user.commercial_partner_id.id != subscription.commercial_partner_id.id
                or authorization.state != 'pending'):
            raise werkzeug.exceptions.Forbidden("This operation can only be done by the customer")
        values = {
            'expire_date': expire_date,
            'state': 'accepted'
        }
        authorization.sudo().write(values)
        database = subscription.sudo().database_ids[0]
        spoof_vals = {
            'ip': request.httprequest.environ['REMOTE_ADDR'],
            'user_id': responsible_user.id,
            'action': "access_granted",
            'reason': authorization.reason,
            'db_id': database.id,
            'db_uuid': database.db_uuid,
            'success': True,
            'restricted': False,  # There are no restrictions for the partner responsible of the subscription
        }
        request.env['openerp.enterprise.spoof'].sudo().create(spoof_vals)
        return request.redirect('/my/subscription/authorizations/{}'.format(subscription.id))

    @http.route('/my/subscription/authorization/<int:auth_id>/extend', type='http', auth="user", methods=['GET'], website=True)
    def extend_authorization(self, auth_id):
        return self._authorization_actions(auth_id, 'access_extended')

    @http.route('/my/subscription/authorization/<int:auth_id>/revoke', type='http', auth="user", methods=['GET'], website=True)
    def revoke_authorization(self, auth_id):
        return self._authorization_actions(auth_id, 'access_revoked')

    @http.route('/my/subscription/authorization/<int:auth_id>/deny', type='http', auth="user", methods=['GET'], website=True)
    def denied_authorization(self, auth_id):
        return self._authorization_actions(auth_id, 'access_denied')

    def _authorization_actions(self, auth_id, action):
        current_user = request.env.user
        authorization = request.env['openerp.enterprise.spoof.authorization'].sudo().browse(auth_id).exists()
        if not authorization:
            return request.render('website.page_404')
        subscription = authorization.subscription_id
        if (request.session.impostor_info
                or subscription.commercial_partner_id.id != current_user.commercial_partner_id.id):
            raise werkzeug.exceptions.Forbidden("This operation can only be done by the customer")

        database = subscription.sudo().database_ids[0]
        spoof_vals = {
            'ip': request.httprequest.environ['REMOTE_ADDR'],
            'user_id': current_user.id,
            'action': action,
            'reason': '',
            'db_id': database.id,
            'db_uuid': database.db_uuid,
            'success': True,
            'restricted': False,  # There are no restrictions for the partner responsible of the subscription
        }
        if authorization.state == 'pending' and action == 'access_denied':
            authorization.write({'expire_date': datetime.date.today(), 'state': 'denied'})
        elif authorization.state == 'accepted' and action == 'access_revoked':
            authorization.write({'state': 'revoked'})
        elif (authorization.state == 'accepted' and action == 'access_extended'
                and authorization.expire_date < datetime.date.today()):
            expire_date = datetime.date.today() + relativedelta(days=7)
            authorization.write({'expire_date': expire_date})
            spoof_vals.update(reason=str(expire_date))
        else:
            raise werkzeug.exceptions.BadRequest()
        request.env['openerp.enterprise.spoof'].sudo().create(spoof_vals)
        return request.redirect("/my/subscription/authorizations/%s" % subscription.id)

    @http.route('/my/subscription/authorizations/<int:subscription_id>', type='http', auth="user", methods=['GET'], website=True, readonly=True)
    def my_authorizations(self, subscription_id):
        subscription = request.env['sale.order'].browse(subscription_id)
        if not subscription.sudo().is_staff_restricted:
            return request.render("website.page_404")
        if not request.session.impostor_info:
            is_responsible = bool(subscription.commercial_partner_id.id == request.env.user.commercial_partner_id.id)
        else:
            # spoofing the responsible isn't allowed
            is_responsible = False
        authorizations = subscription.sudo().spoof_authorization_ids.filtered(lambda s: s.state != 'pending' and s.state != 'draft')
        database_name = subscription.sudo().database_ids[:1].db_name
        return request.render('openerp_enterprise.portal_spoof_authorization_list', {
            'today_date': datetime.date.today(),
            'database_name': database_name,
            'authorizations': authorizations,
            'sale_order': subscription,
            'is_responsible': is_responsible,
        })

    @route('/my/subscription/switch-yearly/<int:order_id>/<access_token>', auth='public', website=True)
    def switch_yearly(self,  order_id, access_token):
        order_sudo, redirection = self._get_subscription(access_token, order_id)
        if redirection:
            return redirection

        if not order_sudo._can_switch_yearly():
            raise werkzeug.exceptions.Forbidden()

        url = order_sudo._portal_switch_yearly_url()
        return request.redirect(url)

class OECustomerPortal(CustomerPortal):
    def _prepare_portal_layout_values(self):
        """ prepare the values to render portal layout """
        values = super()._prepare_portal_layout_values()
        partner = request.env.user.partner_id
        # get customer sales rep
        if partner.user_id:
            sales_rep = partner.user_id
        else:
            sales_rep = False
        values.update({
            'sales_rep': sales_rep,
            'company': request.website.company_id,
            'user': request.env.user
        })
        return values

    @http.route('/my/issues/<int:issue_id>', type='http', methods=['GET'], auth='user', website=True, readonly=True)
    def my_issue_redirect(self, issue_id):
        """ This route redirect the user trying to access their issue.
            Issues were removed on Odoo v11 and have been migrated to tasks.
            All the old issue's ids we 'incremented' by 1,000,000.
            So we check if the id is less than a million, if so we add a million and redirect to my/task"""
        offset = 1000000
        if issue_id < offset:
            issue_id += offset

        return request.redirect("/my/tasks/%s" % issue_id, code=301)

    @http.route([
        '/my/task/<int:task_id>/close',  # keep old route for emails sent in the wild
        '/my/tasks/<int:task_id>/close',
    ], type='http', auth='user', website=True, csrf=False)
    def close_ticket(self, task_id):
        # retrieve the task by Id and in the "Help" project
        task = request.env['project.task'].search([('id', '=', task_id), ('project_id', '=', request.env.ref('openerp_enterprise.project_help').id)])
        if task:
            try:
                task.check_access_rule('write')
                task.sudo().write({'stage_id': request.env.ref('openerp_enterprise.stage_done_help').id})
                task.sudo()._message_log(body="Ticket closed by client: %s" % task.partner_id.name)
            except AccessError:
                pass

        return request.redirect("/my/tasks/%s" % task_id)

    def _get_subinfo_details(self, db_name, db_uuid):
        def _warn(str):
            _logger.warning(str)
            return False
        subinfo = {'sub_plan': 'SubNotFound'}
        OED = request.env['openerp.enterprise.database']
        domain = [('db_name', '=', db_name), ('db_uuid', '=', db_uuid), ('hosting', '=', 'saas')]

        try:
            # If we have access to the subscription (without sudo)
            db = OED.search(domain, limit=1, order='id desc')
            subinfo['sub_link'] = db.subscription_id and db.subscription_id._get_portal_url(access_token=True) or 'nosub'
        except Exception:
            # Else if we are not logged in or don't have access to the subscription with the current user, we don't want
            # to provide an access to the subsciption with an access token.
            db = OED.sudo().search(domain, limit=1, order='id desc')
            subinfo['sub_link'] = False

        if not db:
            return _warn('MySub: db not found')

        if db.hosting != 'saas':
            return _warn('MySub: Wrong hosting')

        subinfo['url'] = db.url
        subinfo['name'] = db.db_name

        if not db.subscription_id or db.subscription_id.subscription_state in ('1_draft', '2_renewed'):
            if db._is_edu():
                subinfo['sub_plan'] = 'Edu'
                subinfo['until'] = db.date_valid.date()
            elif not db.extra_apps:
                subinfo['sub_plan'] = 'OneAppFree'
            else:
                subinfo['sub_plan'] = 'Trial'
                subinfo['expireIn'] = (db.date_valid.date() - fields.Date.today()).days
        elif db.subscription_id.to_renew:
            subinfo['sub_plan'] = 'ToRenew'
            expdate = db.subscription_id.upsell_date or db.subscription_id.next_invoice_date
            auto_close_limit = db.subscription_id.plan_id.auto_close_limit
            expdate = db.subscription_id.upsell_date or (
                db.subscription_id.next_invoice_date + relativedelta(days=auto_close_limit)
            )
            subinfo['expireIn'] = (expdate - fields.Date.today()).days
        elif db.subscription_id.subscription_state in ('3_progress', '4_paused'):
            expdate = db.subscription_id.upsell_date or db.subscription_id.next_invoice_date
            subinfo['sub_plan'] = 'InProgress'
            subinfo['until'] = expdate

        return subinfo

    @http.route(['/my/subscription/status'], type='http', auth='public', website=True, methods=['POST'], csrf=False, readonly=True)
    def subscription_status(self, **kw):
        # This controller is called from customer database to know the status of the subscription linked to this db.
        if request.env.user._is_public():
            db_name, db_uuid = kw.get('db_name'), kw.get('db_uuid')
            if not db_name or not db_uuid:
                return request.redirect('/my')
            subinfo = self._get_subinfo_details(db_name, db_uuid)
            return request.render('openerp_enterprise.subscription_status', {'subscription_info': subinfo})
        else:
            return request.redirect('/my', code=307)

    @http.route(['/my', '/my/home'], csrf=False)
    def home(self, **kw):
        response = super().home(**kw)

        db_name, db_uuid = kw.get('db_name'), kw.get('db_uuid')
        if db_name and db_uuid:
            response.qcontext['subscription_info'] = self._get_subinfo_details(db_name, db_uuid)

        return response

    def details_form_validate(self, data, partner_creation=False):
        error, error_message = super().details_form_validate(data, partner_creation=partner_creation)
        if error.get('name', '') == 'error':
            partner = request.env['res.users'].browse(request.uid).partner_id
            if partner.parent_id.portal_accounting_firm_level and partner.type == 'contact':
                # this is a contact of an accounting firm, allow the user to change it's name
                # if there is not SO or INV linked to it directly.
                has_so = request.env["sale.order"].search([
                    ('partner_id', 'child_of', partner.id),
                    ('state', 'in', ['sent', 'sale'])
                ], limit=1)
                has_invoice = request.env['account.move'].search([
                    ('move_type', 'in', ['out_invoice', 'out_refund']),
                    ('partner_id', 'child_of', partner.id),
                    ('state', '=', 'posted')
                ], limit=1)
                if not has_so and not has_invoice:
                    error.pop('name', None)
                    error_message.remove(
                        _("Changing your name is not allowed once invoices have been issued for your account. "
                          "Please contact us directly for this operation.")
                    )
        return error, error_message

    @http.route(['/my/packs', '/my/packs/page/<int:page>'], type='http', auth="user", website=True, readonly=True)
    def portal_my_packs(self, page=1, sortby=None, **kw):
        # compat. redirect to tasks
        return request.redirect('/my/tasks')

    @route('/my/pack/<int:account_id>', auth='user', type='http', website=True, readonly=True)
    def support_pack(self, account_id, **kwargs):
        # pack ownership is verified when searching orders
        orders = request.env["sale.order"].search([('analytic_account_id', '=', account_id)])
        if not orders:
            raise werkzeug.exceptions.NotFound()
        elif len(orders) == 1:
            return request.redirect(orders[0].access_url)
        else:
            # multiple orders, return list of related tasks
            return request.redirect('/my/tasks?search_in=old_success_pack&search=%d' % account_id)


class OESaleProjectCustomerPortal(SaleProjectCustomerPortal):
    def _task_get_search_domain(self, search_in, search):
        if search_in == "old_success_pack":
            try:
                return [('sale_order_id.analytic_account_id', '=', int(search))]
            except ValueError:
                return [expression.FALSE_LEAF]
        return super()._task_get_search_domain(search_in, search)

    def _prepare_tasks_values(self, *args, **kwargs):
        values = super()._prepare_tasks_values(*args, **kwargs)
        if values.get('search_in') == "old_success_pack":
            # hide search key in UI - user are not expected to search on this
            values.update(search_in='content', search="")
        return values


class LocalizationControllers(http.Controller):

    @from_role('nightly')
    @http.route('/website/dump/translations/nightly', auth='public', readonly=True)
    def dump_translations_nightly(self, lang=False, module='odoo_dot_com', **kw):
        return self._dump_translations(lang=lang, module=module)

    @http.route('/website/dump/translations', auth='user', readonly=True)
    def dump_translations(self, lang=False, module='odoo_dot_com', **kw):
        if not (request.env.user._is_admin() or
                request.env.user.has_group('openerp_enterprise.group_translator')):
            return request.not_found()
        return self._dump_translations(lang=lang, module=module)

    def _dump_translations(self, lang=False, module='odoo_dot_com'):
        # no lang for source terms export
        if lang and not request.env['res.lang'].search_count([('code', '=', lang)]):
            return request.not_found()

        with contextlib.closing(io.BytesIO()) as buf:
            odoo.tools.trans_export(lang, [module], buf, 'po', request.cr)
            out = buf.getvalue()

        headers = [
            ('Content-Type', 'text/plain'),
            ('Content-Length', len(out)),
        ]
        return request.make_response(out, headers)


class OEMailController(MailController):
    @http.route(['/mail/view/subscription/<string:code>'], type='http', auth='user', methods=['GET'])
    def _mail_view_subscription(self, code):
        subscription = request.env['sale.order']._get_most_valid_subscription(code)

        if not subscription:
            raise werkzeug.exceptions.NotFound()

        return self._redirect_to_record(subscription._name, subscription.id)
